@@ -10,21 +10,25 @@
 #
 
 import argparse
+import datetime
+import getpass
 import logging
 import os
 import re
 
 from typing import Optional
-from .. import api, crypto, utils
+from cryptography.hazmat.primitives.serialization import pkcs12
+
+from .. import api, crypto, utils, error
 from ..params import KeeperParams
 
-from .base import GroupCommand, dump_report_data
+from .base import GroupCommand, Command, dump_report_data, report_output_parser, field_to_title
 from .enterprise import EnterpriseCommand
 
 from ..proto import automator_pb2 as automator_proto
 from ..proto import ssocloud_pb2 as ssocloud
 
-automator_list_parser = argparse.ArgumentParser(prog='automator-list')
+automator_list_parser = argparse.ArgumentParser(prog='automator-list', parents=[report_output_parser])
 
 automator_view_parser = argparse.ArgumentParser(prog='automator-view')
 automator_view_parser.add_argument('target', help='Automator ID or Name.')
@@ -38,7 +42,7 @@ automator_edit_parser = argparse.ArgumentParser(prog='automator-edit')
 automator_edit_parser.add_argument('target', help='Automator ID or Name.')
 automator_edit_parser.add_argument('--name', dest='name', action='store', help='Automator name.')
 automator_edit_parser.add_argument('--url', dest='url', action='store', help='Automator Webhook URL.')
-automator_edit_parser.add_argument('--skill', dest='skill', action='append', choices=['device', 'team'],
+automator_edit_parser.add_argument('--skill', dest='skill', action='append', choices=['device', 'team', 'team_for_user'],
                                    help='Automator Skills.')
 automator_edit_parser.add_argument('--set', dest='setting', metavar="KEY=VALUE", action='append',
                                    help='Automator Settings. Use value: '
@@ -54,13 +58,16 @@ automator_delete_parser.add_argument('--force', '-f', dest='force', action='stor
 automator_action_parser = argparse.ArgumentParser(prog='automator-action')
 automator_action_parser.add_argument('target', help='Automator ID or Name.')
 
+automator_certificate_parser = argparse.ArgumentParser(prog='automator-certificate')
+automator_certificate_parser.add_argument('file', help='PFX Certificate file path.')
+
 
 def register_commands(commands):
     commands['automator'] = AutomatorCommand()
 
 
 def register_command_info(_, command_info):
-    command_info['automator'] = 'Manage Automator endpoints'
+    command_info['automator'] = 'Manage Automator endpoints for SSO Cloud'
 
 
 class AutomatorCommand(GroupCommand):
@@ -80,6 +87,7 @@ class AutomatorCommand(GroupCommand):
         action_command = AutomatorLogCommand()
         self.register_command('log', action_command, 'Retrieves automator logs.')
         self.register_command('log-clear', action_command, 'Clears automator logs.')
+        self.register_command('certificate', AutomatorCertificateCommand(), 'Display certificate information.')
 
         self.default_verb = 'list'
 
@@ -91,6 +99,8 @@ class AutomatorMixin(object):
             return 'Device Approval'
         elif skill == automator_proto.TEAM_APPROVAL:
             return 'Team Approval'
+        elif skill == automator_proto.TEAM_FOR_USER_APPROVAL:
+            return 'Team For User Approval'
         else:
             return str(skill)
 
@@ -99,28 +109,69 @@ class AutomatorMixin(object):
         return '{0}={1}'.format(setting.settingTag, setting.settingValue or '')
 
     @staticmethod
-    def dump_automators(params):  # type: (KeeperParams) -> None
+    def dump_automators(params, fmt=None, filename=None):
+        # type: (KeeperParams, Optional[str], Optional[str]) -> None
         table = []
-        headers = ['ID', 'Name', 'Enabled', 'URL', 'Skills']
+        headers = ['id', 'name', 'node_id', 'enabled', 'url', 'skills']
+        if fmt and fmt != 'json':
+            headers = [field_to_title(x) for x in headers]
         if params.automators:
             for info in params.automators:
-                row = [info.automatorId, info.name, info.enabled, info.url,
+                row = [info.automatorId, info.name, info.nodeId, info.enabled, info.url,
                        [AutomatorMixin.skill_to_name(x.skillType) for x in info.automatorSkills]]
                 table.append(row)
-        dump_report_data(table, headers=headers)
+        return dump_report_data(table, headers=headers, fmt=fmt, filename=filename)
 
     @staticmethod
     def dump_automator(endpoint, status=False):    # type: (automator_proto.AutomatorInfo, bool) -> None
         logging.info('{0:>32s}: {1}'.format('Automator ID', endpoint.automatorId))
         logging.info('{0:>32s}: {1}'.format('Name', endpoint.name))
+        logging.info('{0:>32s}: {1}'.format('Node ID', endpoint.nodeId))
         logging.info('{0:>32s}: {1}'.format('URL', endpoint.url))
+        if endpoint.version:
+            logging.info('{0:>32s}: {1}'.format('Version', endpoint.version))
         logging.info('{0:>32s}: {1}'.format('Enabled', 'Yes' if endpoint.enabled else 'No'))
         if status:
             logging.info('{0:>32s}: {1}'.format('Initialized', 'Yes' if endpoint.status.initialized else 'No'))
+            if endpoint.status.initialized:
+                delta = datetime.datetime.now() - datetime.datetime.fromtimestamp (endpoint.status.initializedTimestamp//1000)
+                days = delta.days
+                seconds = delta.seconds
+                hours = seconds // 3600
+                uptime = ''
+                if days > 0:
+                    uptime += f'{days} days '
+                    if hours > 0:
+                        uptime += f'{hours} hours'
+                else:
+                    minutes = (seconds % 3600) // 60
+                    if hours > 0:
+                        uptime += f'{hours} hours'
+                        if minutes > 0:
+                            uptime += f' {minutes} minutes'
+                    else:
+                        uptime += f' {minutes} minutes'
+                logging.info('{0:>32s}: {1}'.format('Uptime', uptime))
+
+            if endpoint.status.sslCertificateExpiration > 0:
+                dt = datetime.datetime.fromtimestamp(endpoint.status.sslCertificateExpiration//1000)
+                logging.info('{0:>32}: {1}'.format('Certificate Expires', dt.strftime('%x')))
+
             if endpoint.status.numberOfDevicesApproved > 0:
                 logging.info('{0:>32}: {1}'.format('Approved Devices', endpoint.status.numberOfDevicesApproved))
             if endpoint.status.numberOfDevicesDenied > 0:
                 logging.info('{0:>32}: {1}'.format('Denied Devices', endpoint.status.numberOfDevicesDenied))
+
+            if endpoint.status.numberOfTeamsApproved > 0:
+                logging.info('{0:>32}: {1}'.format('Approved Teams', endpoint.status.numberOfTeamsApproved))
+            if endpoint.status.numberOfTeamsDenied > 0:
+                logging.info('{0:>32}: {1}'.format('Denied Teams', endpoint.status.numberOfTeamsDenied))
+
+            if endpoint.status.numberOfTeamMembershipsApproved > 0:
+                logging.info('{0:>32}: {1}'.format('Approved Team Memberships', endpoint.status.numberOfTeamMembershipsApproved))
+            if endpoint.status.numberOfTeamMembershipsDenied > 0:
+                logging.info('{0:>32}: {1}'.format('Denied Team Memberships', endpoint.status.numberOfTeamMembershipsDenied))
+
             if endpoint.status.numberOfErrors > 0:
                 logging.info('{0:>32}: {1}'.format('Number of Errors', endpoint.status.numberOfErrors))
 
@@ -190,19 +241,19 @@ class AutomatorMixin(object):
 
 
 class AutomatorListCommand(EnterpriseCommand, AutomatorMixin):
-    def get_parser(self):  # type: () -> Optional[argparse.ArgumentParser]
+    def get_parser(self):
         return automator_list_parser
 
-    def execute(self, params, **kwargs):  # type: (KeeperParams, **any) -> any
+    def execute(self, params, **kwargs):
         self.ensure_loaded(params, True)
-        self.dump_automators(params)
+        return self.dump_automators(params, fmt=kwargs.get('format'), filename=kwargs.get('output'))
 
 
 class AutomatorViewCommand(EnterpriseCommand, AutomatorMixin):
-    def get_parser(self):  # type: () -> Optional[argparse.ArgumentParser]
+    def get_parser(self):
         return automator_view_parser
 
-    def execute(self, params, **kwargs):  # type: (KeeperParams, **any) -> any
+    def execute(self, params, **kwargs):
         automator = self.find_automator(params, kwargs.get('target'))
 
         rq = automator_proto.AdminGetAutomatorRequest()
@@ -233,12 +284,17 @@ class AutomatorCreateCommand(EnterpriseCommand, AutomatorMixin):
         if len(nodes) > 1:
             logging.warning('Node name \'%s\' is not unique. Use Node ID.', node)
             return
-        matched_node = nodes[0]
+        matched_node_id = nodes[0]['node_id']
+        self.ensure_loaded(params, False)
+        # if params.automators:    # type: list[automator_proto.AutomatorInfo]
+        #     n = next((True for x in params.automators if x.nodeId == matched_node_id), None)
+        #     if n:
+        #         logging.warning('Automator for node \"%s\" already exists', node)
+        #         return
 
         rq = automator_proto.AdminCreateAutomatorRequest()
-        rq.nodeId = matched_node['node_id']
+        rq.nodeId = matched_node_id
         rq.name = name
-
         rs = api.communicate_rest(params, rq, 'automator/automator_create', rs_type=automator_proto.AdminResponse)
         if rs.success:
             self.dump_automator(rs.automatorInfo[0])
@@ -257,11 +313,16 @@ class AutomatorEditCommand(EnterpriseCommand, AutomatorMixin):
         automator = self.find_automator(params, kwargs.get('target'))
         rq = automator_proto.AdminEditAutomatorRequest()
         rq.automatorId = automator.automatorId
+        rq.enabled = automator.enabled
         name = kwargs['name']
         if name:
             rq.name = name
-        url = kwargs['url']
+        url = kwargs['url']   # type: str
         if url:
+            if not url.startswith('http'):
+                url = 'https://' + url
+            if url.endswith('/'):
+                url = url.rstrip('/')
             rq.url = url
         skills = kwargs.get('skill')
         if skills:
@@ -270,6 +331,8 @@ class AutomatorEditCommand(EnterpriseCommand, AutomatorMixin):
                     rq.skillTypes.append(automator_proto.DEVICE_APPROVAL)
                 elif skill == 'team':
                     rq.skillTypes.append(automator_proto.TEAM_APPROVAL)
+                elif skill == 'team_for_user':
+                    rq.skillTypes.append(automator_proto.TEAM_FOR_USER_APPROVAL)
                 else:
                     logging.warning('Unsupported skill: \"%s\"', skill)
                     return
@@ -315,7 +378,7 @@ class AutomatorEditCommand(EnterpriseCommand, AutomatorMixin):
                             if parts[0].lower() == 'file':
                                 filename = parts[1].strip()
                                 if os.path.isfile(filename):
-                                    with open(filename, 'r') as f:
+                                    with open(filename, 'r', encoding='utf-8') as f:
                                         value = f.read()
                             if parts[0].lower() == 'base64':
                                 filename = parts[1].strip()
@@ -355,23 +418,25 @@ class AutomatorSetupCommand(EnterpriseCommand, AutomatorMixin):
         if not rs.success:
             logging.warning('Automator \"%s\" setup step #1 error: %s', automator.name, rs.message)
             return
+
         rq = automator_proto.AdminSetupAutomatorRequest()
         rq.automatorId = automator.automatorId
         rq.automatorState = automator_proto.NEEDS_CRYPTO_STEP_2
-        automator_public_key = crypto.load_ec_public_key(rs.automatorEcPublicKey)
+        automator_public_key = crypto.load_ec_public_key(rs.automatorEccPublicKey)
         keys = params.enterprise['keys']
         if 'ecc_encrypted_private_key' in keys:
             encrypted_ec_private_key = utils.base64_url_decode(keys['ecc_encrypted_private_key'])
             ec_private_key = crypto.decrypt_aes_v2(encrypted_ec_private_key, params.enterprise['unencrypted_tree_key'])
             encrypted_ec_private_key = crypto.encrypt_ec(ec_private_key, automator_public_key)
-            rq.encryptedEcEnterprisePrivateKey = encrypted_ec_private_key
+            rq.encryptedEccEnterprisePrivateKey = encrypted_ec_private_key
 
-        if 'rsa_encrypted_private_key' in keys:
+        if not params.forbid_rsa and 'rsa_encrypted_private_key' in keys:
             encrypted_rsa_private_key = utils.base64_url_decode(keys['rsa_encrypted_private_key'])
             rsa_private_key = crypto.decrypt_aes_v2(encrypted_rsa_private_key, params.enterprise['unencrypted_tree_key'])
             encrypted_rsa_private_key = crypto.encrypt_ec(rsa_private_key, automator_public_key)
             rq.encryptedRsaEnterprisePrivateKey = encrypted_rsa_private_key
 
+        rq.encryptedTreeKey = crypto.encrypt_ec(params.enterprise['unencrypted_tree_key'], automator_public_key)
         rs = api.communicate_rest(params, rq, 'automator/automator_setup',
                                   rs_type=automator_proto.AdminSetupAutomatorResponse)
         if rs.success:
@@ -441,4 +506,49 @@ class AutomatorLogCommand(EnterpriseCommand, AutomatorMixin):
         rq.automatorId = automator.automatorId
         rs = api.communicate_rest(params, rq, endpoint, rs_type=automator_proto.AdminResponse)
         if rs.automatorInfo:
-            self.dump_automator(rs.automatorInfo[0])
+            endpoint = rs.automatorInfo[0]
+            if endpoint.logEntries:
+                logging.info('\n{0:>32s}\n'.format('Automator Log'))
+                for log in endpoint.logEntries:
+                    logging.info('<{0}> {1} - {2}'.format(log.messageLevel, log.serverTime, log.message))
+            else:
+                logging.info('\nNo log entries found.')
+
+
+class AutomatorCertificateCommand(Command):
+    def get_parser(self):
+        return automator_certificate_parser
+
+    def execute(self, params, **kwargs):
+        filename = kwargs.get('file')
+        if not filename:
+            raise error.CommandError('automator-certificate', 'PFX file name parameter is required')
+
+        filepath = os.path.expanduser(filename)
+        if not os.path.isfile(filepath):
+            raise error.CommandError('automator-certificate', f'File \"{filename}\" cannot be found')
+
+        password = getpass.getpass(prompt='...' + 'Certificate Passphrase'.rjust(30) + ': ', stream=None)
+        if not password:
+            password = None
+
+        with open(filepath, 'rb') as pfx:
+            _, cert, chain = pkcs12.load_key_and_certificates(pfx.read(), password.encode('utf-8'))
+
+        def print_cert(title, c):
+            print('')
+            print(f'{"":>16} {title}')
+            print(f'{"Subject:":>16} {c.subject.rfc4514_string()}')
+            print(f'{"Issuer:":>16} {c.issuer.rfc4514_string()}')
+            print(f'{"Valid from:":>16} {c.not_valid_before}')
+            print(f'{"Valid until:":>16} {c.not_valid_after}')
+
+        print_cert('MAIN CERTIFICATE', cert)
+        if chain:
+            for i in range(len(chain)):
+                print_cert(f'ADDITIONAL CERTIFICATE #{i+1}', chain[i])
+
+    def is_authorised(self):
+        return False
+
+    
\ No newline at end of file
