@@ -16,13 +16,14 @@ import json
 import logging
 import os
 import re
+import shlex
 import shutil
-import tempfile
-from typing import Optional, Callable, Iterator, List, Iterable, Tuple
-
+import subprocess
 import sys
+import tempfile
+from typing import Optional, Callable, List, Iterable, Tuple
 
-from .base import Command, RecordMixin, dump_report_data
+from .base import Command, RecordMixin, dump_report_data, field_to_title, report_output_parser
 from .record import find_record, RecordListCommand
 from .ssh_agent import add_ssh_key, try_extract_private_key, SshAgentCommand
 from ..attachment import prepare_attachment_download
@@ -31,19 +32,24 @@ from ..params import KeeperParams
 from ..subfolder import find_folders, get_folder_path, try_resolve_path
 from ..vault import TypedRecord, KeeperRecord, PasswordRecord
 
-ssh_parser = argparse.ArgumentParser(prog='ssh',
-                                     description='Establishes connection to external server using SSH. ')
+ssh_parser = argparse.ArgumentParser(prog='ssh', description='Establishes connection to external server using SSH.',
+                                     parents=[report_output_parser])
+ssh_parser.add_argument('-d', '--destination', action='store', metavar='LOGIN@HOST[:PORT]',
+                        help='SSH endpoint')
 ssh_parser.add_argument('record', nargs='?', type=str, action='store',
                         help='record path or UID. Record types: "SSH Key", "Server"')
-ssh_parser.add_argument('destination', nargs='?', type=str, action='store',
-                        metavar='LOGIN@HOST[:PORT]', help='Optional. SSH endpoint')
+ssh_parser.add_argument('command', nargs='*', type=str, action='store',
+                        help='Remote command')
 
-mysql_parser = argparse.ArgumentParser(prog='mysql', description='Establishes connection to MySQL server.')
+mysql_parser = argparse.ArgumentParser(prog='mysql', description='Establishes connection to MySQL server.',
+                                       parents=[report_output_parser])
 mysql_parser.add_argument('record', nargs='?', type=str, action='store',
                           help='record path or UID. Record types: "Database"')
+mysql_parser.add_argument('query', nargs='*', type=str, action='store',
+                          help='SQL query')
 
-postgres_parser = argparse.ArgumentParser(prog='postgresql',
-                                          description='Establishes connection to Postgres/Redshift servers.')
+postgres_parser = argparse.ArgumentParser(prog='postgresql', description='Establishes connection to Postgres/Redshift servers.',
+                                          parents=[report_output_parser])
 postgres_parser.add_argument('record', nargs='?', type=str, action='store',
                              help='record path or UID. Record types: "Database"')
 postgres_parser.add_argument('database', nargs='?', type=str, action='store',
@@ -55,15 +61,18 @@ rdp_parser.add_argument('record', nargs='?', type=str, action='store',
                         help='record path or UID. Record types: "Server"')
 
 
-connect_parser = argparse.ArgumentParser(prog='connect', description='Establishes connection to external server')
+connect_parser = argparse.ArgumentParser(prog='connect', description='Establishes connection to external server.',
+                                         parents=[report_output_parser])
 connect_parser.add_argument('--syntax-help', dest='syntax_help', action='store_true',
                             help='display help on command format and template parameters')
-connect_parser.add_argument('-n', '--new', dest='new_data', action='store_true', help='request per-user data')
 connect_parser.add_argument('-s', '--sort', dest='sort_by', action='store', choices=['endpoint', 'title', 'folder'],
                             help='sort output')
+connect_parser.add_argument('-n', '--new', dest='new_data', action='store_true', help='request per-user data')
 connect_parser.add_argument('-f', '--filter', dest='filter_by', action='store', help='filter output')
 connect_parser.add_argument('endpoint', nargs='?', action='store', type=str,
                             help='endpoint name or full record path to endpoint')
+connect_parser.add_argument('parameters', nargs='*', type=str, action='store',
+                            help='Command parameters')
 
 mysql = ''
 postgresql = ''
@@ -118,21 +127,23 @@ class BaseConnectCommand(Command, RecordMixin):
     def support_extra_parameters(self):
         return True
 
-    SHELL_SUBSTITUTION = {
-        '`': r'\`',
-        '$': r'\$',
-        '?': r'\?',
-        '*': r'\*',
-        '^': r'\^',
-        '(': r'\(',
-        ')': r'\)'
-    }
+    # SHELL_SUBSTITUTION = {
+    #     '`': r'\`',
+    #     '$': r'\$',
+    #     '?': r'\?',
+    #     '*': r'\*',
+    #     '^': r'\^',
+    #     '(': r'\(',
+    #     ')': r'\)'
+    # }
 
     def execute_shell(self):
-        logging.debug('Executing "%s" ...', self.command)
+        logging.debug('Executing "%s"', self.command)
         try:
-            command = self.command.translate(str.maketrans(BaseConnectCommand.SHELL_SUBSTITUTION))
-            os.system(command)
+            # command = self.command.translate(str.maketrans(BaseConnectCommand.SHELL_SUBSTITUTION))
+            command = self.command
+            subprocess.run(shlex.split(command))
+            # os.system(command)
         finally:
             self.command = ''
             for cb in self.run_at_the_end:
@@ -161,28 +172,6 @@ class BaseConnectCommand(Command, RecordMixin):
             options += f' {self.extra_parameters}'
         return options
 
-    @staticmethod
-    def get_record(params, record, types):  # type: (KeeperParams, str, Iterator[str]) -> Optional[TypedRecord]
-        if not record:
-            ls = RecordListCommand()
-            ls.execute(params, record_type=types, verbose=True)
-            return
-
-        try:
-            record = find_record(params, record, types)
-        except Exception as e:
-            logging.warning(e)
-            return
-
-        if not isinstance(record, TypedRecord):
-            logging.warning('Only typed records are supported')
-            return
-
-        if record.record_type not in types:
-            logging.warning('Command supports %s records only', ' and '.join(types))
-            return
-        return record
-
     @staticmethod
     def get_parameter_value(params, record, parameter, temp_files, **kwargs):
         # type: (KeeperParams, KeeperRecord, str, list, ...) -> Optional[str]
@@ -228,11 +217,11 @@ class ConnectSshCommand(BaseConnectCommand):
         return ssh_parser
 
     def execute(self, params, **kwargs):
+        ssh_record_types = ['serverCredentials', 'sshKeys', 'pamMachine', 'pamUser']
         record_name = kwargs['record'] if 'record' in kwargs else None
         if not record_name:
             ls = RecordListCommand()
-            ls.execute(params, record_type=['serverCredentials', 'sshKeys'], verbose=True)
-            return
+            return ls.execute(params, record_type=ssh_record_types, verbose=True, **kwargs)
 
         record = None     # type: Optional[KeeperRecord]
         if record_name in params.record_cache:
@@ -253,7 +242,7 @@ class ConnectSshCommand(BaseConnectCommand):
 
         if record is None:
             ls = RecordListCommand()
-            result = ls.execute(params, record_type=['serverCredentials', 'sshKeys'], format='json', verbose=True)
+            result = ls.execute(params, record_type=ssh_record_types, format='json', verbose=True)
             if result:
                 try:
                     recs = json.loads(result)
@@ -310,10 +299,13 @@ class ConnectSshCommand(BaseConnectCommand):
 
         self.run_at_the_end.clear()
 
+        command = ['ssh']
         options = self.get_extra_options(params, record, 'ssh')
-        self.command = f'ssh{options} {login}@{host_name}'
+        if options:
+            command.append(shlex.split(options.strip(), posix=False))
+        command.append(f'{login}@{host_name}')
         if port:
-            self.command += f' -p {port}'
+            command.extend(['-p', port])
 
         pk = try_extract_private_key(params, record)
         if pk:
@@ -325,14 +317,14 @@ class ConnectSshCommand(BaseConnectCommand):
             password = BaseConnectCommand.get_record_field(record, 'password')
             if password:
                 if shutil.which('sshpass'):
-                    self.command = 'sshpass -e ' + self.command
+                    command = ['sshpass', '-e'] + command
                     os.putenv('SSHPASS', password)
 
                     def clear_env():
                         os.putenv('SSHPASS', '')
                     self.run_at_the_end.append(clear_env)
                 else:
-                    self.command += ' -o PubkeyAuthentication=no'
+                    command.extend(['-o', 'PubkeyAuthentication=no'])
                     try:
                         import pyperclip
                         pyperclip.copy(password)
@@ -347,6 +339,12 @@ class ConnectSshCommand(BaseConnectCommand):
                         logging.debug(e)
                         logging.info('Failed to copy password to clipboard')
 
+        cmd = kwargs.get('command')
+        if isinstance(cmd, list) and len(cmd) > 0:
+            command.append('--')
+            command.extend(cmd)
+
+        self.command = shlex.join(command)
         logging.info('Connecting to "%s" ...', record.title)
         self.execute_shell()
 
@@ -356,10 +354,12 @@ class ConnectMysqlCommand(BaseConnectCommand):
         return mysql_parser
 
     def execute(self, params, **kwargs):
-        name = kwargs['record'] if 'record' in kwargs else None
-        record = self.get_record(params, name, ['databaseCredentials', 'serverCredentials'])
-        if not record:
-            return
+        record_name = kwargs.pop('record', None)
+        if not record_name:
+            ls = RecordListCommand()
+            return ls.execute(params, record_type=['databaseCredentials', 'serverCredentials'], verbose=True, **kwargs)
+
+        record = find_record(params, record_name, types=['databaseCredentials', 'serverCredentials'])
 
         login = BaseConnectCommand.get_record_field(record, 'login')
         if not login:
@@ -374,11 +374,13 @@ class ConnectMysqlCommand(BaseConnectCommand):
 
         self.run_at_the_end.clear()
 
+        command = ['mysql']
         options = self.get_extra_options(params, record, 'mysql')
-        self.command = f'mysql{options}'
-        self.command += f' --host {host_name} --user {login}'
+        if options:
+            command.append(options.strip())
+        command.extend(['--host', host_name, '--user', login])
         if port:
-            self.command += f' --port {port}'
+            command.extend(['--port', port])
 
         password = BaseConnectCommand.get_record_field(record, 'password')
         if password:
@@ -388,6 +390,11 @@ class ConnectMysqlCommand(BaseConnectCommand):
                 os.putenv('MYSQL_PWD', '')
             self.run_at_the_end.append(clear_env)
 
+        query = kwargs.get('query')
+        if isinstance(query, list) and len(query) > 0:
+            command.extend(['--execute', shlex.join(query)])
+
+        self.command = shlex.join(command)
         logging.info('Connecting to "%s" ...', record.title)
         self.execute_shell()
 
@@ -397,10 +404,12 @@ class ConnectPostgresCommand(BaseConnectCommand):
         return postgres_parser
 
     def execute(self, params, **kwargs):
-        name = kwargs['record'] if 'record' in kwargs else None
-        record = self.get_record(params, name, ['databaseCredentials', 'serverCredentials'])
-        if not record:
-            return
+        record_name = kwargs.pop('record', None)
+        if not record_name:
+            ls = RecordListCommand()
+            return ls.execute(params, record_type=['databaseCredentials', 'serverCredentials'], verbose=True, **kwargs)
+
+        record = find_record(params, record_name, types=['databaseCredentials', 'serverCredentials'])
 
         login = BaseConnectCommand.get_record_field(record, 'login')
         if not login:
@@ -419,10 +428,13 @@ class ConnectPostgresCommand(BaseConnectCommand):
             database = 'template1'
             logging.info(f'\nConnecting to the default database: {database}\n')
 
-        self.command = f'{postgresql} {self.extra_parameters} -h {host_name}'
+        command = [postgresql]    # type: List[str]
+        if self.extra_parameters:
+            command.extend(self.extra_parameters)
+        command.extend(['-h', host_name])
         if port:
-            self.command += f' -p {port}'
-        self.command += f' -U {login} -w {database}'
+            command.extend(['-p', port])
+        command.extend(['-U', login, '-w', database])
         self.run_at_the_end.clear()
 
         password = BaseConnectCommand.get_record_field(record, 'password')
@@ -433,6 +445,7 @@ class ConnectPostgresCommand(BaseConnectCommand):
                 os.putenv('PGPASSWORD', '')
             self.run_at_the_end.append(clear_env)
 
+        self.command = shlex.join(command)
         logging.info('Connecting to "%s" ...', record.title)
         self.execute_shell()
 
@@ -442,10 +455,12 @@ class ConnectRdpCommand(BaseConnectCommand):
         return rdp_parser
 
     def execute(self, params, **kwargs):
-        name = kwargs['record'] if 'record' in kwargs else None
-        record = self.get_record(params, name, ['serverCredentials'])
-        if not record:
-            return
+        record_name = kwargs.pop('record', None)
+        if not record_name:
+            ls = RecordListCommand()
+            return ls.execute(params, record_type=['serverCredentials'], verbose=True, **kwargs)
+
+        record = find_record(params, record_name, types=['serverCredentials'])
 
         login = BaseConnectCommand.get_record_field(record, 'login')
         if not login:
@@ -460,15 +475,20 @@ class ConnectRdpCommand(BaseConnectCommand):
 
         password = BaseConnectCommand.get_record_field(record, 'password')
         if password:
-            os.system(f'cmdkey /generic:{host_name} /user:{login} /pass:{password} > NUL')
+            command = ['cmdkey', f'/generic:{host_name}', f'/user:{login}', f'/pass:{password}']
+            subprocess.run(command, stdout=subprocess.DEVNULL)
+            # os.system(f'cmdkey /generic:{host_name} /user:{login} /pass:{password} > NUL')
 
             def clear_password():
-                os.system(f'cmdkey /delete:{host_name} > NUL')
+                subprocess.run(['cmdkey', f'/delete:{host_name}'], stdout=subprocess.DEVNULL)
+                # os.system(f'cmdkey /delete:{host_name} > NUL')
             self.run_at_the_end.append(clear_password)
 
-        self.command = f'mstsc /v:{host_name}'
+        options = f'/v:{host_name}'
         if port:
-            self.command += ':' + port
+            options += ':' + port
+        command = ['mstsc', options]
+        self.command = shlex.join(command)
 
         logging.info('Connecting to "%s" ...', record.title)
         self.execute_shell()
@@ -533,13 +553,13 @@ class ConnectCommand(BaseConnectCommand):
         return connect_parser
 
     def execute(self, params, **kwargs):
-        if kwargs.get('syntax_help'):
+        if kwargs.pop('syntax_help', None):
             logging.info(connect_command_description)
             return
 
         ConnectCommand.find_endpoints(params)
 
-        endpoint = kwargs.get('endpoint')
+        endpoint = kwargs.pop('endpoint', None)
         if endpoint:
             endpoints = [x for x in ConnectCommand.Endpoints if x.name == endpoint]
             if not endpoints:
@@ -574,7 +594,7 @@ class ConnectCommand(BaseConnectCommand):
                 if len(endpoints) == 1:
                     record = KeeperRecord.load(params, endpoints[0].record_uid)
                     if record:
-                        self.connect_endpoint(params, endpoints[0].name, record)
+                        self.connect_endpoint(params, endpoints[0].name, record, **kwargs)
                 else:
                     logging.warning("Connect endpoint '%s' is not unique", endpoint)
                     ConnectCommand.dump_endpoints(endpoints)
@@ -585,33 +605,38 @@ class ConnectCommand(BaseConnectCommand):
                 logging.info("Connect endpoint '%s' not found", endpoint)
         else:
             if ConnectCommand.Endpoints:
-                sorted_by = kwargs.get('sort_by') or 'endpoint'
-                filter_by = kwargs.get('filter_by') or ''
-                logging.info("Available connect endpoints")
-                if filter_by:
-                    logging.info('Filtered by "%s"', filter_by)
-                    filter_by = filter_by.lower()
-                ConnectCommand.dump_endpoints(ConnectCommand.Endpoints, filter_by, sorted_by)
+                return ConnectCommand.dump_endpoints(ConnectCommand.Endpoints, **kwargs)
             else:
                 logging.info("No connect endpoints found")
-            return
 
     @staticmethod
-    def dump_endpoints(endpoints, filter_by='', sorted_by=''):
-        logging.info('')
-        headers = ['Endpoint', 'Description', 'Record Title', 'Folder(s)']
+    def dump_endpoints(endpoints, **kwargs):   # type: (List[ConnectEndpoint], ...) -> Optional[str]
+        sort_by = kwargs.get('sort_by') or ''
+        filter_by = kwargs.get('filter_by') or ''
+        logging.info("Available connect endpoints")
+        if filter_by:
+            logging.info('Filtered by "%s"', filter_by)
+            filter_by = filter_by.lower()
+        fmt = kwargs.get('format', '')
+        verbose = fmt != 'table'
+
+        headers = ['endpoint', 'description', 'record_title', 'record_uid', 'folders']
+        if fmt != 'json':
+            headers = [field_to_title(x) for x in headers]
+
         table = []
         for endpoint in endpoints:
             title = endpoint.record_title
-            folder = endpoint.paths[0] if len(endpoint.paths) > 0 else '/'
+            folder = endpoint.paths[0] if len(endpoint.paths) > 0 else ''
             if filter_by:
-                if not any([x for x in [endpoint.name.lower(), title.lower(), folder.lower()] if x.find(filter_by) >= 0]):
+                if not any([x for x in [endpoint.name.lower(), title.lower(), endpoint.record_uid, folder.lower()] if x.find(filter_by) >= 0]):
                     continue
-            if len(title) > 23:
+            if not verbose and len(title) > 23:
                 title = title[:20] + '...'
-            table.append([endpoint.name, endpoint.description or '', title, folder])
-        table.sort(key=lambda x: x[3] if sorted_by == 'folder' else x[2] if sorted_by == 'title' else x[0])
-        dump_report_data(table, headers, row_number=True)
+            table.append([endpoint.name, endpoint.description or '', title, endpoint.record_uid, folder])
+
+        sorted_by = 4 if sort_by == 'folder' else 2 if sort_by == 'title' else 0
+        return dump_report_data(table, headers, fmt=fmt, output=kwargs.get('output'), row_number=True, sort_by=sorted_by)
 
     @staticmethod
     def find_endpoints(params):   # type: (KeeperParams) -> None
@@ -665,7 +690,6 @@ class ConnectCommand(BaseConnectCommand):
 
     @staticmethod
     def add_ssh_keys(params, endpoint, record, temp_files):
-        added = set()
         # type: (KeeperParams, str, KeeperRecord, List[str]) -> Iterable[Callable]
         pk = try_extract_private_key(params, record)
         if pk:
@@ -720,25 +744,33 @@ class ConnectCommand(BaseConnectCommand):
                     os.putenv(key_name, '')
                 yield clear_env
 
-    def connect_endpoint(self, params, endpoint, record):
-        # type: (KeeperParams, str, KeeperRecord) -> None
+    def connect_endpoint(self, params, endpoint, record, **kwargs):
+        # type: (KeeperParams, str, KeeperRecord, ...) -> None
         temp_files = []
         try:
             command = BaseConnectCommand.get_custom_field(record, f'connect:{endpoint}:pre')
             if command:
                 command = BaseConnectCommand.get_command_string(params, record, command, temp_files, endpoint=endpoint)
                 if command:
-                    os.system(command)
+                    args = shlex.split(command)
+                    subprocess.run(args)
+                    # os.system(command)
 
             command = ConnectCommand.get_custom_field(record, f'connect:{endpoint}')
             if command:
-                self.command = ConnectCommand.get_command_string(params, record, command, temp_files, endpoint=endpoint)
-                if self.command:
+                cmd = ConnectCommand.get_command_string(params, record, command, temp_files, endpoint=endpoint)
+                if cmd:
                     self.run_at_the_end.extend(
                         ConnectCommand.add_ssh_keys(params, endpoint, record, temp_files))
                     self.run_at_the_end.extend(
                         ConnectCommand.add_environment_variables(params, endpoint, record, temp_files))
 
+                    command = shlex.split(cmd, posix=False)
+                    parameters = kwargs.get('parameters')
+                    if isinstance(parameters, list) and len(parameters) > 0:
+                        command.extend(parameters)
+
+                    self.command = shlex.join(command)
                     logging.info('Connecting to "%s" ...', record.title)
                     self.execute_shell()
 
@@ -746,7 +778,9 @@ class ConnectCommand(BaseConnectCommand):
             if command:
                 command = ConnectCommand.get_command_string(params, record, command, temp_files, endpoint=endpoint)
                 if command:
-                    os.system(command)
+                    args = shlex.split(command)
+                    subprocess.run(args)
+                    # os.system(command)
         finally:
             for file in temp_files:
                 os.remove(file)
