@@ -15,7 +15,7 @@ import json
 import logging
 import re
 from collections import OrderedDict
-from typing import Optional, Tuple
+from typing import Optional, Tuple, Dict
 
 from ..utils import is_url, is_email
 from .base import raise_parse_exception, suppress_exit, Command
@@ -36,7 +36,7 @@ def register_command_info(aliases, command_info):
 
 convert_parser = argparse.ArgumentParser(prog='convert', description='Convert record(s) to use record types')
 convert_parser.add_argument(
-    '-t', '--record_type', dest='record_type', action='store', help='Convert to record type'
+    '-t', '--record-type', '--record_type', dest='record_type', action='store', help='Convert to record type'
 )
 convert_parser.add_argument(
     '-q', '--quiet', dest='quiet', action='store_true', help="Don't display info about records matched and converted"
@@ -70,10 +70,10 @@ def get_matching_records_from_folder(params, folder_uid, regex, url_regex, attac
     if folder_uid in params.subfolder_record_cache:
         for uid in params.subfolder_record_cache[folder_uid]:
             if not ignore_owner:
-                if uid not in params.meta_data_cache:
+                if uid not in params.record_owner_cache:
                     continue
-                md = params.meta_data_cache[uid]
-                if not md.get('owner', False):
+                own = params.record_owner_cache[uid]
+                if not own.owner is True:
                     continue
             if uid not in params.record_cache:
                 continue
@@ -148,7 +148,7 @@ class ConvertCommand(Command):
             return
 
         folder = params.folder_cache.get(params.current_folder, params.root_folder)
-        records_by_folder = {}
+        records_by_folder = {}   # type: Dict
         folder_path = {}
         for pattern in pattern_list:
             if pattern in params.record_cache:
@@ -190,20 +190,12 @@ class ConvertCommand(Command):
         dry_run = kwargs.get('dry_run', False)
         if dry_run:
             print(
-                f'The following {len(record_uids)} records that you own were matched'
-                f' and would be converted to records with type "login":'
+                f'The following {len(record_uids)} record(s) that you own were matched'
+                f' and would be converted to records with type "{record_type}":'
             )
 
             print('\n'.join(f' {k}  {v}' for k, v in record_names.items()))
         else:
-            rq = {
-                'command': 'sync_down',
-                'revision': 0,
-                'include': api.NON_SHARED_DATA_SCOPE + api.EXPLICIT
-            }
-            rs = api.communicate(params, rq)
-            nsd = {x['record_uid']: x['data'] for x in rs.get('non_shared_data', [])}
-
             records = []
             for record_uid in record_uids:
                 convert_result = ConvertCommand.convert_to_record_type_data(record_uid, params, type_info)
@@ -251,15 +243,6 @@ class ConvertCommand(Command):
                         rc.record_file.append(rf)
                 rc.data = crypto.encrypt_aes_v2(api.get_record_data_json_bytes(v3_data), record_key)
 
-                # Non shared data
-                if record_uid in nsd:
-                    try:
-                        non_shared_data = utils.base64_url_decode(nsd[record_uid])
-                        non_shared_data = crypto.decrypt_aes_v1(non_shared_data, params.data_key)
-                        rc.non_shared_data = crypto.encrypt_aes_v2(non_shared_data, params.data_key)
-                    except Exception as e:
-                        logging.debug('Non shared data conversion failed for record %s: ', record_uid, e)
-
                 # Get share folder of the record so that we can convert the Record Folder Key
                 shared_folders = find_parent_top_folder(params, record_uid)
 
