@@ -12,21 +12,13 @@
 import base64
 import collections
 import logging
+from typing import Dict
 
 from .base import Command, user_choice
-from .. import api, utils, crypto, rest_api
+from .. import api, utils, crypto
 from ..error import CommandError
-from ..params import KeeperParams
-from ..proto.enterprise_pb2 import RoleTeam, RoleTeams
-
-
-def decrypt_role_key(params, role_key):
-    if role_key['key_type'] == 'encrypted_by_data_key':
-        return api.decrypt_data(role_key['encrypted_key'], params.data_key)
-    elif role_key['key_type'] == 'encrypted_by_public_key':
-        return api.decrypt_rsa(role_key['encrypted_key'], params.rsa_key)
-    else:
-        return None
+from ..params import KeeperParams, PublicKeys
+from ..proto import enterprise_pb2
 
 
 class EnterpriseCommand(Command):
@@ -34,7 +26,7 @@ class EnterpriseCommand(Command):
         super(EnterpriseCommand, self).__init__()
         self.public_keys = {}
         self.team_keys = {}
-        self.nodes = None
+        self._node_map = None
 
     def execute_args(self, params, args, **kwargs):
         if params.enterprise:
@@ -42,73 +34,6 @@ class EnterpriseCommand(Command):
         else:
             raise CommandError('', 'This command  is only available for Administrators of Keeper.')
 
-    def get_public_keys(self, params, emails):
-        # type: (EnterpriseCommand, KeeperParams, dict) -> None
-
-        for email in emails:
-            emails[email] = self.public_keys.get(email.lower())
-
-        email_list = [x[0] for x in emails.items() if x[1] is None]
-        if len(email_list) == 0:
-            return
-
-        rq = {
-            'command': 'public_keys',
-            'key_owners': email_list
-        }
-        rs = api.communicate(params, rq)
-        for pko in rs['public_keys']:
-            if 'public_key' in pko:
-                public_key = crypto.load_rsa_public_key(utils.base64_url_decode(pko['public_key']))
-                self.public_keys[pko['key_owner'].lower()] = public_key
-                emails[pko['key_owner']] = public_key
-
-    def get_public_key(self, params, email):
-        # type: (EnterpriseCommand, KeeperParams, str) -> any
-
-        public_key = self.public_keys.get(email.lower())
-        if public_key is None:
-            emails = {
-                email: None
-            }
-            self.get_public_keys(params, emails)
-            public_key = emails[email]
-
-        return public_key
-
-    def get_team_key(self, params, team_uid):
-        team_key = self.team_keys.get(team_uid)
-        if team_key is None:
-            if 'teams' in params.enterprise:
-                for team in params.enterprise['teams']:
-                    if team['team_uid'] == team_uid:
-                        if 'encrypted_team_key' in team:
-                            enc_team_key = team['encrypted_team_key']  # type: str
-                            team_key = crypto.decrypt_aes_v2(utils.base64_url_decode(enc_team_key), params.enterprise['unencrypted_tree_key'])
-                        break
-
-        if team_key is None:
-            rq = {
-                'command': 'team_get_keys',
-                'teams': [team_uid]
-            }
-            rs = api.communicate(params, rq)
-            if rs['result'] == 'success':
-                ko = rs['keys'][0]
-                if 'key' in ko:
-                    encrypted_key = utils.base64_url_decode(ko['key'])
-                    key_type = ko['type']
-                    if key_type == 1:
-                        team_key = crypto.decrypt_aes_v1(encrypted_key, params.data_key)
-                    elif key_type == 2:
-                        team_key = crypto.decrypt_rsa(encrypted_key, params.rsa_key)
-                    elif key_type == 3:
-                        team_key = crypto.decrypt_aes_v2(encrypted_key, params.data_key)
-
-        if team_key is not None:
-            self.team_keys[team_uid] = team_key
-        return team_key
-
     def get_role_users_change_batch(self, params, roles, add_user, remove_user, force=False):
         """Get batch of requests for changing enterprise role users"""
         request_batch = []
@@ -127,7 +52,7 @@ class EnterpriseCommand(Command):
                     else:
                         logging.warning('User %s could be resolved', u)
 
-        user_pkeys = {}
+        user_pkeys = {}    # type: Dict[str, PublicKeys]
         for role in roles:
             role_id = role['role_id']
             for user_id in user_changes:
@@ -145,13 +70,13 @@ class EnterpriseCommand(Command):
                     role_key = next((rk2['role_key'] for rk2 in role_keys2 if rk2['role_id'] == role_id), None)
                     if role_key:
                         encrypted_key_decoded = base64.urlsafe_b64decode(role_key + '==')
-                        role_key = rest_api.decrypt_aes(
+                        role_key = crypto.decrypt_aes_v2(
                             encrypted_key_decoded, params.enterprise['unencrypted_tree_key']
                         )
                     else:
                         role_keys = params.enterprise.get('role_keys', [])
                         role_key = next((
-                            decrypt_role_key(params, rk) for rk in role_keys if rk['role_id'] == role_id
+                            self.decrypt_role_key(params, rk) for rk in role_keys if rk['role_id'] == role_id
                         ), None)
                 rq = {
                     'command': 'role_user_add' if is_add else 'role_user_remove',
@@ -162,19 +87,38 @@ class EnterpriseCommand(Command):
                     if user_id not in user_pkeys:
                         answer = 'y' if force else user_choice(
                             'Do you want to grant administrative privileges to {0}'.format(email), 'yn', 'n')
-                        public_key = None
                         if answer == 'y':
-                            public_key = self.get_public_key(params, email)
-                            if public_key is None:
+                            api.load_user_public_keys(params, [email], False)
+                            user_keys = params.key_cache.get(email)
+                            if user_keys:
+                                user_pkeys[user_id] = params.key_cache[email]
+                            else:
                                 logging.warning('Cannot get public key for user %s', email)
-                        user_pkeys[user_id] = public_key
-                    if user_pkeys[user_id]:
-                        encrypted_tree_key = crypto.encrypt_rsa(params.enterprise['unencrypted_tree_key'],
-                                                                user_pkeys[user_id])
-                        rq['tree_key'] = utils.base64_url_encode(encrypted_tree_key)
-                        if role_key:
-                            encrypted_role_key = crypto.encrypt_rsa(role_key, user_pkeys[user_id])
-                            rq['role_admin_key'] = utils.base64_url_encode(encrypted_role_key)
+
+                    public_key = user_pkeys.get(user_id)
+                    if public_key:
+                        tree_key = params.enterprise['unencrypted_tree_key']
+                        if params.forbid_rsa and public_key.ec:
+                            ec_key = crypto.load_ec_public_key(public_key.ec)
+                            encrypted_tree_key = crypto.encrypt_ec(tree_key, ec_key)
+                            rq['tree_key'] = utils.base64_url_encode(encrypted_tree_key)
+                            rq['tree_key_type'] = 'encrypted_by_public_key_ecc'
+                            if role_key:
+                                encrypted_role_key = crypto.encrypt_ec(role_key, ec_key)
+                                rq['role_admin_key'] = utils.base64_url_encode(encrypted_role_key)
+                                rq['role_admin_key_type'] = 'encrypted_by_public_key_ecc'
+                        elif not params.forbid_rsa and public_key.rsa:
+                            rsa_key = crypto.load_rsa_public_key(public_key.rsa)
+                            encrypted_tree_key = crypto.encrypt_rsa(tree_key, rsa_key)
+                            rq['tree_key'] = utils.base64_url_encode(encrypted_tree_key)
+                            rq['tree_key_type'] = 'encrypted_by_public_key'
+                            if role_key:
+                                encrypted_role_key = crypto.encrypt_rsa(role_key, rsa_key)
+                                rq['role_admin_key'] = utils.base64_url_encode(encrypted_role_key)
+                                rq['role_admin_key_type'] = 'encrypted_by_public_key'
+                        else:
+                            continue
+
                         request_batch.append(rq)
                 else:
                     request_batch.append(rq)
@@ -182,10 +126,11 @@ class EnterpriseCommand(Command):
 
     @staticmethod
     def decrypt_role_key(params, rk):
+        encrypted_key = utils.base64_url_decode(rk['encrypted_key'])
         if rk['key_type'] == 'encrypted_by_data_key':
-            return api.decrypt_data(rk['encrypted_key'], params.data_key)
+            return crypto.decrypt_aes_v1(encrypted_key, params.data_key)
         elif rk['key_type'] == 'encrypted_by_public_key':
-            return api.decrypt_rsa(rk['encrypted_key'], params.rsa_key)
+            return crypto.decrypt_rsa(encrypted_key, params.rsa_key2)
         else:
             return None
 
@@ -200,9 +145,9 @@ class EnterpriseCommand(Command):
             team_list = add_team if is_add else remove_team
             if team_list:
                 if is_add:
-                    add_teams = RoleTeams()
+                    add_teams = enterprise_pb2.RoleTeams()
                 else:
-                    remove_teams = RoleTeams()
+                    remove_teams = enterprise_pb2.RoleTeams()
                 for team in team_list:
                     team_node = next((
                         t for t in params.enterprise.get('teams', []) if team in (t['team_uid'], t['name'])
@@ -226,7 +171,7 @@ class EnterpriseCommand(Command):
                 if is_managed_role:
                     logging.warning('Teams cannot be assigned to roles with administrative permissions.')
                 else:
-                    role_team = RoleTeam()
+                    role_team = enterprise_pb2.RoleTeam()
                     role_team.role_id = role_id
                     role_team.teamUid = utils.base64_url_decode(team_id)
                     role_name = role['data']['displayname']
@@ -252,9 +197,9 @@ class EnterpriseCommand(Command):
             role_list = add_roles if is_add else remove_roles
             if role_list:
                 if is_add:
-                    add_role_teams = RoleTeams()
+                    add_role_teams = enterprise_pb2.RoleTeams()
                 else:
-                    remove_role_teams = RoleTeams()
+                    remove_role_teams = enterprise_pb2.RoleTeams()
                 for role in role_list:
                     role_node = next((
                         r for r in params.enterprise['roles']
@@ -289,7 +234,7 @@ class EnterpriseCommand(Command):
                                 'Team %s is not in "%s" role: Remove from role is skipped', team['name'], role_name
                             )
                         else:
-                            role_team = RoleTeam()
+                            role_team = enterprise_pb2.RoleTeam()
                             role_team.role_id = role_id
                             role_team.teamUid = utils.base64_url_decode(team['team_uid'])
                             team_name = team['name']
@@ -315,16 +260,48 @@ class EnterpriseCommand(Command):
         if rs['result'] == 'success':
             return rs['base_id']
 
-    def get_node_path(self, params, node_id):
-        if self.nodes is None:
-            self.nodes = {}
-        for node in params.enterprise['nodes']:
-            self.nodes[node['node_id']] = (node['data'].get('displayname') or params.enterprise['enterprise_name'], node.get('parent_id') or 0)
+    @staticmethod
+    def get_enterprise_ids(params, num_ids=1):
+        enterprise_ids = []
+        if num_ids < 1:
+            return enterprise_ids
+        rqs = []
+        rq_max_ids = 100
+        num_requested = num_ids
+        while num_requested:
+            batch_num_ids = min(num_ids, rq_max_ids)
+            num_requested -= batch_num_ids
+            rq = {
+                'command': 'enterprise_allocate_ids',
+                'number_requested': batch_num_ids
+            }
+            rqs.append(rq)
+        if rqs:
+            rss = api.execute_batch(params, rqs) if num_ids > rq_max_ids else [api.communicate(params, rqs[0])]
+            for i, rs in enumerate(rss):
+                if rs.get('result') == 'success':
+                    base_id = rs.get('base_id')
+                    number_allocated = rs.get('number_allocated')
+                    batch_ids = [base_id + i for i in range(number_allocated)]
+                    enterprise_ids.extend(batch_ids)
+                else:
+                    # Batch of new enterprise ids could not be allocated (
+                    rq = rqs[i]
+                    enterprise_ids.extend([None for _ in range(rq.get('number_requested', 0))])
+        return enterprise_ids
+
+    def get_node_path(self, params, node_id, omit_root=False):
+        if self._node_map is None:
+            self._node_map = {
+                x['node_id']: (x['data'].get('displayname') or x['name'] or str(x['node_id']) if x.get('parent_id', 0) > 0 else params.enterprise['enterprise_name'], x.get('parent_id', 0))
+                for x in params.enterprise['nodes']}
         path = ''
-        node = self.nodes.get(node_id)
+        node = self._node_map.get(node_id)
         while node:
+            if omit_root and node[1] == 0:
+                break
             path = '{0}{1}{2}'.format(node[0], '\\' if path else '', path)
-            node = self.nodes.get(node[1])
+            node = self._node_map.get(node[1])
         return path
 
     @staticmethod
