@@ -25,17 +25,14 @@ from cryptography.hazmat.backends import default_backend
 from cryptography.hazmat.primitives import serialization, hashes
 from cryptography.hazmat.primitives.asymmetric import rsa, ec, ed25519, padding, utils
 
-from .base import GroupCommand, Command, dump_report_data, user_choice
+from .base import GroupCommand, Command, dump_report_data, user_choice, report_output_parser
 from .. import vault, attachment
 from ..params import KeeperParams
 
 ssh_agent_info_parser = argparse.ArgumentParser(prog='ssh-agent info', description='Display ssh agent status')
 ssh_agent_start_parser = argparse.ArgumentParser(prog='ssh-agent start', description='Start ssh agent')
 ssh_agent_stop_parser = argparse.ArgumentParser(prog='ssh-agent stop', description='Stop ssh agent')
-
-ssh_agent_log_parser = argparse.ArgumentParser(prog='ssh-agent log', description='Display ssh agent logs')
-ssh_agent_log_parser.add_argument('--format', dest='format', action='store', choices=['csv', 'json', 'table'], default='table', help='output format.')
-ssh_agent_log_parser.add_argument('--output', dest='output', action='store', help='output file name. (ignored for table format)')
+ssh_agent_log_parser = argparse.ArgumentParser(prog='ssh-agent log', description='Display ssh agent logs', parents=[report_output_parser])
 
 
 class ConnectSshAgent:
@@ -181,6 +178,16 @@ def is_private_key_name(name):     # type: (str) -> bool
 
     return False
 
+KEY_SIZE_MIN = 119 # Smallest possible size for ed25519 private key in PKCS#8 format
+KEY_SIZE_MAX = 4000
+def is_valid_key_value(value):
+    return isinstance(value, str) and KEY_SIZE_MIN <= len(value) < KEY_SIZE_MAX
+
+def is_valid_key_file(file):
+    try:
+        return KEY_SIZE_MIN <= file.size < KEY_SIZE_MAX
+    except:
+        return False
 
 def try_extract_private_key(params, record_or_uid):
     # type: (KeeperParams, Union[str, vault.KeeperRecord]) -> Optional[Tuple[str, str]]
@@ -207,7 +214,7 @@ def try_extract_private_key(params, record_or_uid):
     # check notes field
     if not private_key:
         if isinstance(record, (vault.PasswordRecord, vault.TypedRecord)):
-            if 444 < len(record.notes) < 4000:
+            if is_valid_key_value(record.notes):
                 header, _, _ = record.notes.partition('\n')
                 if is_private_key(header):
                     private_key = record.notes
@@ -217,14 +224,14 @@ def try_extract_private_key(params, record_or_uid):
         if isinstance(record, vault.TypedRecord):
             try_values = (x.get_default_value() for x in itertools.chain(record.fields, record.custom) if x.type in ('text', 'multiline', 'secret', 'note'))
             for value in (x for x in try_values if x):
-                if isinstance(value, str) and 444 < len(value) < 4000:
+                if is_valid_key_value(value):
                     header, _, _ = value.partition('\n')
                     if is_private_key(header):
                         private_key = value
                         break
         elif isinstance(record, vault.PasswordRecord):
             for value in (x.value for x in record.custom if x.value):
-                if isinstance(value, str) and 444 < len(value) < 4000:
+                if is_valid_key_value(value):
                     header, _, _ = value.partition('\n')
                     if is_private_key(header):
                         private_key = value
@@ -244,7 +251,7 @@ def try_extract_private_key(params, record_or_uid):
                         if file_record.name and file_record.name != file_record.title:
                             names.append(file_record.name)
                         if any(True for x in names if is_private_key_name(x)):
-                            if 444 < file_record.size < 4000:
+                            if is_valid_key_file(file_record):
                                 key_file_uids.append(file_uid)
                 if len(key_file_uids) == 1:
                     download_rq = next(attachment.prepare_attachment_download(params, key_file_uids[0]), None)
@@ -258,7 +265,7 @@ def try_extract_private_key(params, record_or_uid):
                     if atta.name and atta.title != atta.name:
                         names.append(atta.name)
                     if any(True for x in names if is_private_key_name(x)):
-                        if 444 < atta.size < 4000:
+                        if is_valid_key_file(atta):
                             key_attachment_ids.append(atta.id)
             if len(key_attachment_ids) == 1:
                 download_rq = next(attachment.prepare_attachment_download(params, record.record_uid, key_attachment_ids[0]), None)
@@ -308,7 +315,7 @@ def add_ssh_key(private_key_pem, passphrase, key_name):   # type: (str, str, str
         remove_payload += ssh_agent_encode_long(public_numbers.n)
         remove_payload = SSH_AGENTC_REMOVE_IDENTITY.to_bytes(1, byteorder='big') + ssh_agent_encode_bytes(remove_payload)
     elif isinstance(private_key, ec.EllipticCurvePrivateKey):
-        curve_name = 'nistp381' if private_key.curve.name == 'secp384r1' else \
+        curve_name = 'nistp384' if private_key.curve.name == 'secp384r1' else \
             'nistp521' if private_key.curve.name == 'secp521r1' else \
                 'nistp256' if private_key.curve.name == 'secp256r1' else ''
         if not curve_name:
@@ -504,7 +511,7 @@ class SshAgentContext(logging.Handler):
                             hash_algorithm = hashes.SHA512()
                             signature_method = 'rsa-sha2-512'
                         elif flags & SSH_AGENT_RSA_SHA2_256:
-                            hash_algorithm = hashes.SHA512()
+                            hash_algorithm = hashes.SHA256()
                             signature_method = 'rsa-sha2-256'
                         else:
                             hash_algorithm = hashes.SHA1()
@@ -755,7 +762,7 @@ class SshAgentCommand(GroupCommand):
         self.default_verb = 'info'
 
     def is_authorised(self):
-        return False
+        return True
 
 
 class SshAgentInfoCommand(Command):
