@@ -12,7 +12,7 @@
 import abc
 import json
 import logging
-from typing import Optional, List, Set, Tuple
+from typing import Optional, List, Set, Tuple, Dict
 
 from google.protobuf import message
 
@@ -21,9 +21,9 @@ from .proto import enterprise_pb2 as proto
 from . import api, utils, crypto
 
 
-def query_enterprise(params):  # type: (KeeperParams) -> None
+def query_enterprise(params, tree_key=None):  # type: (KeeperParams, Optional[bytes]) -> None
     if not params.enterprise_loader:
-        params.enterprise_loader = _EnterpriseLoader()
+        params.enterprise_loader = _EnterpriseLoader(tree_key)
     params.enterprise_loader.load(params)
 
 
@@ -64,9 +64,10 @@ class EnterpriseInfo(object):
 
 
 class _EnterpriseLoader(object):
-    def __init__(self):
+    def __init__(self, tree_key=None):
         super(_EnterpriseLoader, self).__init__()
         self._enterprise = EnterpriseInfo()
+        self._enterprise._tree_key = tree_key
         self._continuationToken = b''
         self._data_types = {   # type: dict[int, _EnterpriseDataParser]
             proto.NODES: _EnterpriseNodeEntity(self._enterprise),
@@ -88,22 +89,25 @@ class _EnterpriseLoader(object):
             proto.ROLE_ENFORCEMENTS: _EnterpriseRoleEnforcements(self._enterprise),
             proto.MANAGED_COMPANIES: _EnterpriseManagedCompanyEntity(self._enterprise),
             proto.DEVICES_REQUEST_FOR_ADMIN_APPROVAL: _EnterpriseAdminApprovalRequestEntity(self._enterprise),
+            proto.USER_ALIASES: _EnterpriseUserAliasEntity(self._enterprise),
         }
         teams = self._data_types[proto.TEAMS]
         if isinstance(teams, _EnterpriseEntity):
             teams.register_link('team_uid', self._data_types[proto.TEAM_USERS])
             teams.register_link('team_uid', self._data_types[proto.ROLE_TEAMS])
+            teams.register_link('team_uid', self._data_types[proto.QUEUED_TEAM_USERS])
 
         users = self._data_types[proto.USERS]
         if isinstance(teams, _EnterpriseEntity):
             users.register_link('enterprise_user_id', self._data_types[proto.TEAM_USERS])
             users.register_link('enterprise_user_id', self._data_types[proto.ROLE_USERS])
+            teams.register_link('enterprise_user_id', self._data_types[proto.QUEUED_TEAM_USERS])
 
         roles = self._data_types[proto.ROLES]
         if isinstance(roles, _EnterpriseEntity):
-            users.register_link('role_id', self._data_types[proto.ROLE_TEAMS])
-            users.register_link('role_id', self._data_types[proto.ROLE_USERS])
-            users.register_link('role_id', self._data_types[proto.MANAGED_NODES])
+            roles.register_link('role_id', self._data_types[proto.ROLE_TEAMS])
+            roles.register_link('role_id', self._data_types[proto.ROLE_USERS])
+            roles.register_link('role_id', self._data_types[proto.MANAGED_NODES])
 
     @property
     def enterprise(self):
@@ -114,51 +118,86 @@ class _EnterpriseLoader(object):
             params.enterprise = {}
             self._continuationToken = b''
 
-        if not self._enterprise.tree_key:
+        if 'unencrypted_tree_key' not in params.enterprise or 'keys' not in params.enterprise:
             rq = proto.GetEnterpriseDataKeysRequest()
             rs = api.communicate_rest(params, rq, 'enterprise/get_enterprise_data_keys',
                                       rs_type=proto.GetEnterpriseDataKeysResponse)
-            if rs.treeKey:
-                encrypted_tree_key = utils.base64_url_decode(rs.treeKey.treeKey)
-                if rs.treeKey.keyTypeId == proto.ENCRYPTED_BY_DATA_KEY:
-                    self._enterprise._tree_key = crypto.decrypt_aes_v1(encrypted_tree_key, params.data_key)
-                elif rs.treeKey.keyTypeId == proto.ENCRYPTED_BY_PUBLIC_KEY:
-                    if len(encrypted_tree_key) == 60:
-                        self._enterprise._tree_key = crypto.decrypt_aes_v2(encrypted_tree_key, params.msp_tree_key)
-                    else:
-                        self._enterprise._tree_key = api.decrypt_rsa(rs.treeKey.treeKey, params.rsa_key)
-                params.enterprise['unencrypted_tree_key'] = self._enterprise.tree_key
+            if not self._enterprise.tree_key:
+                if rs.treeKey.treeKey:
+                    encrypted_tree_key = utils.base64_url_decode(rs.treeKey.treeKey)
+                    if rs.treeKey.keyTypeId == proto.ENCRYPTED_BY_DATA_KEY:
+                        self._enterprise._tree_key = crypto.decrypt_aes_v1(encrypted_tree_key, params.data_key)
+                    elif rs.treeKey.keyTypeId == proto.ENCRYPTED_BY_PUBLIC_KEY:
+                        self._enterprise._tree_key = crypto.decrypt_rsa(encrypted_tree_key, params.rsa_key2)
+                    elif rs.treeKey.keyTypeId == proto.ENCRYPTED_BY_DATA_KEY_GCM:
+                        self._enterprise._tree_key = crypto.decrypt_aes_v2(encrypted_tree_key, params.data_key)
+                    elif rs.treeKey.keyTypeId == proto.ENCRYPTED_BY_PUBLIC_KEY_ECC:
+                        self._enterprise._tree_key = crypto.decrypt_ec(encrypted_tree_key, params.ecc_key)
+            params.enterprise['unencrypted_tree_key'] = self._enterprise.tree_key
 
+            keys = {}    # type: Dict[str, str]
             if rs.enterpriseKeys:
-                keys = {}
                 if rs.enterpriseKeys.rsaEncryptedPrivateKey:
-                    self._enterprise._rsa_key = \
-                        api.decrypt_aes_plain(rs.enterpriseKeys.rsaEncryptedPrivateKey, self._enterprise.tree_key)
-                    keys['rsa_public_key'] = utils.base64_url_encode(rs.enterpriseKeys.rsaPublicKey)
-                    keys['rsa_encrypted_private_key'] = \
-                        utils.base64_url_encode(rs.enterpriseKeys.rsaEncryptedPrivateKey)
+                    try:
+                        self._enterprise._rsa_key = \
+                            crypto.decrypt_aes_v2(rs.enterpriseKeys.rsaEncryptedPrivateKey, self._enterprise.tree_key)
+                        keys['rsa_public_key'] = utils.base64_url_encode(rs.enterpriseKeys.rsaPublicKey)
+                        keys['rsa_encrypted_private_key'] = \
+                            utils.base64_url_encode(rs.enterpriseKeys.rsaEncryptedPrivateKey)
+                    except:
+                        if not params.forbid_rsa:
+                            logging.info('Error decrypting enterprise RSA key')
+                        keys['rsa_encrypted_private_key'] = ''
                 if rs.enterpriseKeys.eccEncryptedPrivateKey:
-                    self._enterprise._ec_key = \
-                        api.decrypt_aes_plain(rs.enterpriseKeys.eccEncryptedPrivateKey, self._enterprise.tree_key)
                     keys['ecc_public_key'] = utils.base64_url_encode(rs.enterpriseKeys.eccPublicKey)
                     keys['ecc_encrypted_private_key'] = \
                         utils.base64_url_encode(rs.enterpriseKeys.eccEncryptedPrivateKey)
-                params.enterprise['keys'] = keys
+            if 'rsa_encrypted_private_key' not in keys and not params.forbid_rsa:
+                rsa_private, rsa_public = crypto.generate_rsa_key()
+                rsa_private_key = crypto.unload_rsa_private_key(rsa_private)
+                rsa_encrypted_private_key = crypto.encrypt_aes_v2(rsa_private_key, self._enterprise.tree_key)
+                rsa_public_key = crypto.unload_rsa_public_key(rsa_public)
+                rq = proto.EnterpriseKeyPairRequest()
+                rq.enterprisePublicKey = rsa_public_key
+                rq.encryptedEnterprisePrivateKey = rsa_encrypted_private_key
+                rq.keyType = proto.RSA
+                api.communicate_rest(params, rq, 'enterprise/set_enterprise_key_pair')
+                self._enterprise._rsa_key = rsa_private_key
+                keys['rsa_public_key'] = utils.base64_url_encode(rsa_public_key)
+                keys['rsa_encrypted_private_key'] = utils.base64_url_encode(rsa_encrypted_private_key)
+
+            if 'ecc_encrypted_private_key' not in keys:
+                ec_private, ec_public = crypto.generate_ec_key()
+                ec_private_key = crypto.unload_ec_private_key(ec_private)
+                ec_encrypted_private_key = crypto.encrypt_aes_v2(ec_private_key, self._enterprise.tree_key)
+                ec_public_key = crypto.unload_ec_public_key(ec_public)
+                rq = proto.EnterpriseKeyPairRequest()
+                rq.enterprisePublicKey = ec_public_key
+                rq.encryptedEnterprisePrivateKey = ec_encrypted_private_key
+                rq.keyType = proto.ECC
+                api.communicate_rest(params, rq, 'enterprise/set_enterprise_key_pair')
+                keys['ecc_public_key'] = utils.base64_url_encode(ec_public_key)
+                keys['ecc_encrypted_private_key'] = utils.base64_url_encode(ec_encrypted_private_key)
+
+            params.enterprise['keys'] = keys
         entities = set()
         while True:
             rq = proto.EnterpriseDataRequest()
             if self._continuationToken:
                 rq.continuationToken = self._continuationToken
             rs = api.communicate_rest(params, rq, 'enterprise/get_enterprise_data_for_user',
-                                      rs_type=proto.EnterpriseDataResponse)
+                                      rs_type=proto.EnterpriseDataResponse, payload_version=2)
 
             if rs.cacheStatus == proto.CLEAR:
                 for d in self._data_types.values():
                     d.clear(params)
+                self._enterprise._enterprise_name = ''
 
             if not self._enterprise.enterprise_name and rs.generalData:
                 self._enterprise._enterprise_name = rs.generalData.enterpriseName
                 params.enterprise['enterprise_name'] = self._enterprise.enterprise_name
+                if rs.generalData.distributor:
+                    params.enterprise['distributor'] = True
 
             for ed in rs.data:
                 entities.add(ed.entity)
@@ -203,7 +242,7 @@ class _EnterpriseLoader(object):
                 for rk1 in rs.roleKey:
                     params.enterprise['role_keys'].append({
                         'role_id': rk1.roleId,
-                        'encrypted_key': utils.base64_url_encode(rk1.encryptedKey),
+                        'encrypted_key': rk1.encryptedKey,
                         'key_type': _to_key_type(rk1.keyType)
                     })
 
@@ -254,7 +293,7 @@ class _EnterpriseDataParser(abc.ABC):
 class _EnterpriseEntity(_EnterpriseDataParser):
     def __init__(self, enterprise):  # type: (EnterpriseInfo) -> None
         super(_EnterpriseEntity, self).__init__(enterprise)
-        self._links = []     # type: List[Tuple[str, _EnterpriseLink]]
+        self._links = []     # type: List[Tuple[str, _CascadeDeleteLink]]
 
     @abc.abstractmethod
     def get_keeper_entity_id(self, proto_entity):  # type: (dict) -> any
@@ -265,14 +304,14 @@ class _EnterpriseEntity(_EnterpriseDataParser):
         pass
 
     @staticmethod
-    def fix_data(d):
+    def fix_data(d):  # type: (bytes) -> bytes
         idx = d.rfind(b'}')
         if idx < len(d) - 1:
             d = d[:idx+1]
         return d
 
     def register_link(self, keeper_entity_id_name, parser):  # type: (str, _EnterpriseDataParser) -> None
-        if isinstance(parser, _EnterpriseLink):
+        if isinstance(parser, _CascadeDeleteLink):
             self._links.append((keeper_entity_id_name, parser))
 
     def parse(self, params, enterprise_data, **kwargs):  # type: (KeeperParams, proto.EnterpriseData, dict) -> None
@@ -305,7 +344,13 @@ class _EnterpriseEntity(_EnterpriseDataParser):
                 link.cascade_delete(params, keeper_entity_id_name, deleted_entities)
 
 
-class _EnterpriseLink(_EnterpriseDataParser):
+class _CascadeDeleteLink:
+    @abc.abstractmethod
+    def cascade_delete(self, params, keeper_entity_id, deleted_entities):   # type: (KeeperParams, str, Set) -> None
+        pass
+
+
+class _EnterpriseLink(_EnterpriseDataParser, _CascadeDeleteLink):
     @abc.abstractmethod
     def get_keeper_entity1_id(self, proto_entity):  # type: (dict) -> any
         pass
@@ -391,11 +436,12 @@ class _EnterpriseNodeEntity(_EnterpriseEntity):
         data = {}
         if 'encrypted_data' in keeper_entity:
             try:
-                data_json = api.decrypt_data(keeper_entity['encrypted_data'], self.enterprise.tree_key)
+                encrypted_data = utils.base64_url_decode(keeper_entity['encrypted_data'])
+                data_json = crypto.decrypt_aes_v1(encrypted_data, self.enterprise.tree_key)
                 data_json = self.fix_data(data_json)
                 data.update(json.loads(data_json.decode('utf-8')))
             except Exception as e:
-                logging.warning('Decrypt encryption data error: %s', e)
+                logging.info('Decrypt encrypted data error: %s', e)
         elif 'parent_id' not in keeper_entity:
             data['displayname'] = self.enterprise.enterprise_name
         keeper_entity['data'] = data
@@ -419,7 +465,7 @@ class _EnterpriseUserEntity(_EnterpriseEntity):
         _set_or_remove(keeper_entity, 'node_id', proto_entity.nodeId)
         _set_or_remove(keeper_entity, 'username', proto_entity.username)
         _set_or_remove(keeper_entity, 'encrypted_data', proto_entity.encryptedData)
-        _set_or_remove(keeper_entity, 'key_type', proto_entity.keyType)
+        _set_or_remove(keeper_entity, 'key_type', proto_entity.keyType.lower())
         _set_or_remove(keeper_entity, 'status', proto_entity.status)
         _set_or_remove(keeper_entity, 'lock', proto_entity.lock)
         _set_or_remove(keeper_entity, 'user_id', proto_entity.userId)
@@ -427,18 +473,32 @@ class _EnterpriseUserEntity(_EnterpriseEntity):
                        proto_entity.accountShareExpiration if proto_entity.accountShareExpiration > 0 else None)
         _set_or_remove(keeper_entity, 'full_name', proto_entity.fullName if proto_entity.fullName else None)
         _set_or_remove(keeper_entity, 'job_title', proto_entity.jobTitle if proto_entity.jobTitle else None)
+        _set_or_remove(keeper_entity, 'tfa_enabled', proto_entity.tfaEnabled)
+        _set_or_remove(keeper_entity, 'transfer_acceptance_status', proto_entity.transferAcceptanceStatus)
         data = {}
         encrypted_data = keeper_entity.get('encrypted_data')
         if encrypted_data:
-            if keeper_entity.get('key_type') == 'no_key':
-                data['displayname'] = encrypted_data
-            else:
-                try:
-                    data_json = api.decrypt_data(encrypted_data, self.enterprise.tree_key)
+            try:
+                if keeper_entity.get('key_type') == 'no_key':
+                    data['displayname'] = encrypted_data
+                else:
+                    encrypted_data = utils.base64_url_decode(keeper_entity['encrypted_data'])
+                    if keeper_entity.get('key_type') == 'encrypted_by_data_key':
+                        data_json = crypto.decrypt_aes_v1(encrypted_data, self.enterprise.tree_key)
+                    elif keeper_entity['key_type'] == 'encrypted_by_public_key':
+                        rsa_key = crypto.load_rsa_private_key(self.enterprise.rsa_key)
+                        data_json = crypto.decrypt_rsa(encrypted_data, rsa_key)
+                    elif keeper_entity.get('key_type') == 'encrypted_by_data_key_gcm':
+                        data_json = crypto.decrypt_aes_v2(encrypted_data, self.enterprise.tree_key)
+                    elif keeper_entity['key_type'] == 'encrypted_by_public_key_ecc':
+                        ec_key = crypto.load_ec_private_key(self.enterprise.ec_key)
+                        data_json = crypto.decrypt_ec(encrypted_data, ec_key)
+                    else:
+                        raise Exception(f'unsupported key type: {proto_entity.keyType}')
                     data_json = self.fix_data(data_json)
                     data.update(json.loads(data_json.decode('utf-8')))
-                except Exception as e:
-                    logging.warning('Decrypt User data error: %s', e)
+            except Exception as e:
+                logging.warning('Decrypt User data error: %s', e)
         elif 'full_name' in keeper_entity:
             data['displayname'] = keeper_entity['full_name']
         keeper_entity['data'] = data
@@ -487,15 +547,32 @@ class _EnterpriseRoleEntity(_EnterpriseEntity):
         _set_or_remove(keeper_entity, 'encrypted_data', proto_entity.encryptedData)
         _set_or_remove(keeper_entity, 'visible_below', proto_entity.visibleBelow)
         _set_or_remove(keeper_entity, 'new_user_inherit', proto_entity.newUserInherit)
-        _set_or_remove(keeper_entity, 'key_type', proto_entity.keyType)
+        _set_or_remove(keeper_entity, 'key_type', proto_entity.keyType.lower())
         _set_or_remove(keeper_entity, 'role_type', proto_entity.roleType)
         data = {}
         encrypted_data = keeper_entity.get('encrypted_data')
         if encrypted_data:
             try:
-                data_json = api.decrypt_data(encrypted_data, self.enterprise.tree_key)
-                data_json = self.fix_data(data_json)
-                data.update(json.loads(data_json.decode('utf-8')))
+                if keeper_entity.get('key_type') == 'no_key':
+                    data['displayname'] = encrypted_data
+                else:
+                    encrypted_data = utils.base64_url_decode(keeper_entity['encrypted_data'])
+                    if keeper_entity.get('key_type') == 'encrypted_by_data_key':
+                        data_json = crypto.decrypt_aes_v1(encrypted_data, self.enterprise.tree_key)
+                    elif keeper_entity.get('key_type') == 'encrypted_by_public_key':
+                        rsa_key = crypto.load_rsa_private_key(self.enterprise.rsa_key)
+                        data_json = crypto.decrypt_rsa(encrypted_data, rsa_key)
+                    elif keeper_entity.get('key_type') == 'encrypted_by_data_key_gcm':
+                        data_json = crypto.decrypt_aes_v2(encrypted_data, self.enterprise.tree_key)
+                    elif keeper_entity.get('key_type') == 'encrypted_by_public_key_ecc':
+                        ec_key = crypto.load_ec_private_key(self.enterprise.ec_key)
+                        data_json = crypto.decrypt_ec(encrypted_data, ec_key)
+                    else:
+                        raise Exception(f'unsupported key type: {proto_entity.keyType}')
+                    data_json = self.fix_data(data_json)
+                    data.update(json.loads(data_json.decode('utf-8')))
+                    if proto_entity.roleType == "pool_manager":
+                        data['displayname'] = 'MSP Subscription Manager'
             except Exception as e:
                 logging.warning('Decrypt encryption data error: %s', e)
         keeper_entity['data'] = data
@@ -515,8 +592,6 @@ class _EnterpriseRoleEntity(_EnterpriseEntity):
 
 class _EnterpriseLicenseEntity(_EnterpriseEntity):
     def to_keeper_entity(self, proto_entity, keeper_entity):  # type: (proto.License, dict) -> None
-        _set_or_remove(keeper_entity, 'enterprise_license_id', proto_entity.enterpriseLicenseId)
-        _set_or_remove(keeper_entity, 'name', proto_entity.name)
         _set_or_remove(keeper_entity, 'paid', proto_entity.paid)
         _set_or_remove(keeper_entity, 'number_of_seats', proto_entity.numberOfSeats)
         _set_or_remove(keeper_entity, 'expiration', proto_entity.expiration)
@@ -524,6 +599,8 @@ class _EnterpriseLicenseEntity(_EnterpriseEntity):
                        proto_entity.licenseKeyId if proto_entity.licenseKeyId > 0 else None)
         _set_or_remove(keeper_entity, 'product_type_id',
                        proto_entity.productTypeId if proto_entity.productTypeId > 0 else None)
+        _set_or_remove(keeper_entity, 'name', proto_entity.name)
+        _set_or_remove(keeper_entity, 'enterprise_license_id', proto_entity.enterpriseLicenseId)
         _set_or_remove(keeper_entity, 'seats_allocated', proto_entity.seatsAllocated)
         _set_or_remove(keeper_entity, 'seats_pending', proto_entity.seatsPending)
         _set_or_remove(keeper_entity, 'tier', proto_entity.tier)
@@ -534,7 +611,8 @@ class _EnterpriseLicenseEntity(_EnterpriseEntity):
         _set_or_remove(keeper_entity, 'storage_expiration',
                        proto_entity.storageExpiration if proto_entity.storageExpiration > 0 else None)
         _set_or_remove(keeper_entity, 'lic_status', proto_entity.licenseStatus)
-        msp_pool = None
+        _set_or_remove(keeper_entity, 'distributor', proto_entity.distributor)
+
         if proto_entity.mspPool:
             msp_pool = [{
                 'product_id': x.productId,
@@ -542,12 +620,14 @@ class _EnterpriseLicenseEntity(_EnterpriseEntity):
                 'availableSeats': x.availableSeats,
                 'stash': x.stash
             } for x in proto_entity.mspPool]
-        _set_or_remove(keeper_entity, 'msp_pool', msp_pool)
+            _set_or_remove(keeper_entity, 'msp_pool', msp_pool)
+
         if proto_entity.managedBy and proto_entity.managedBy.enterpriseId > 0:
             _set_or_remove(keeper_entity, 'managed_by', {
                 'enterprise_id': proto_entity.managedBy.enterpriseId,
                 'enterprise_name': proto_entity.managedBy.enterpriseName,
             })
+
         if proto_entity.addOns:
             _set_or_remove(keeper_entity, 'add_ons', [{
                 'name': x.name,
@@ -555,7 +635,24 @@ class _EnterpriseLicenseEntity(_EnterpriseEntity):
                 'is_trial': x.isTrial,
                 'created': x.created,
                 'expiration': x.expiration,
+                'included_in_product': x.includedInProduct,
+                'seats': x.seats,
+                'api_call_count': x.apiCallCount,
             } for x in proto_entity.addOns])
+
+        if proto_entity.mspPermits.restricted:
+            _set_or_remove(keeper_entity, 'msp_permits', {
+                'allow_unlimited_licenses': proto_entity.mspPermits.allowUnlimitedLicenses,
+                'allowed_mc_products': [x for x in proto_entity.mspPermits.allowedMcProducts],
+                'allowed_add_ons': [x for x in proto_entity.mspPermits.allowedAddOns],
+                'max_file_plan_type': proto_entity.mspPermits.maxFilePlanType,
+                'mc_defaults': [{
+                    'mc_product': x.mcProduct,
+                    'add_ons': [a for a in x.addOns],
+                    'file_plan_type': x.filePlanType,
+                } for x in proto_entity.mspPermits.mcDefaults]
+            })
+
         _set_or_remove(keeper_entity, 'next_billing_date',
                        proto_entity.nextBillingDate if proto_entity.nextBillingDate > 0 else None)
 
@@ -783,14 +880,16 @@ class _EnterpriseManagedCompanyEntity(_EnterpriseEntity):
         _set_or_remove(keeper_entity, 'tree_key', proto_entity.treeKey if proto_entity.treeKey else None)
         _set_or_remove(keeper_entity, 'tree_key_role', proto_entity.tree_key_role)
         _set_or_remove(keeper_entity, 'file_plan_type', proto_entity.filePlanType)
-        if proto_entity.addOns:
-            _set_or_remove(keeper_entity, 'add_ons', [{
-                'name': x.name,
-                'enabled': x.enabled,
-                'is_trial': x.isTrial,
-                'created': x.created,
-                'expiration': x.expiration,
-            } for x in proto_entity.addOns])
+        _set_or_remove(keeper_entity, 'add_ons', [{
+            'name': x.name,
+            'seats': x.seats,
+            'enabled': x.enabled,
+            'is_trial': x.isTrial,
+            'created': x.created,
+            'expiration': x.expiration,
+            'activation_time': x.activationTime,
+            'included_in_product': x.includedInProduct,
+        } for x in proto_entity.addOns])
 
     def get_keeper_entity_id(self, entity):  # type: (dict) -> any
         return entity.get('mc_enterprise_id')
@@ -805,7 +904,7 @@ class _EnterpriseManagedCompanyEntity(_EnterpriseEntity):
         return 'managed_companies'
 
 
-class _EnterpriseQueuedTeamUserEntity(_EnterpriseDataParser):
+class _EnterpriseQueuedTeamUserEntity(_EnterpriseDataParser, _CascadeDeleteLink):
     def parse(self, params, enterprise_data, **kwargs):  # type: (KeeperParams, proto.EnterpriseData, dict) -> None
         entities = self.get_entities(params)
         entity_map = {x['team_uid']: x for x in entities}
@@ -841,6 +940,40 @@ class _EnterpriseQueuedTeamUserEntity(_EnterpriseDataParser):
     def to_keeper_entity(self, proto_entity, keeper_entity):
         pass
 
+    def cascade_delete(self, params, keeper_entity_id, deleted_entities):   # type: (KeeperParams, str, Set) -> None
+        entities = self.get_entities(params)
+        if not entities:
+            return
+        if keeper_entity_id == 'team_uid':
+            to_keep = [x for x in entities if x['team_uid'] not in deleted_entities]
+            if len(to_keep) < len(entities):
+                entities.clear()
+                entities.extend(to_keep)
+        elif keeper_entity_id == 'enterprise_user_id':
+            for entity in entities:
+                users = entity.get('users')
+                if isinstance(users, set):
+                    users.difference_update(deleted_entities)
+
+
+class _EnterpriseUserAliasEntity(_EnterpriseEntity):
+    def to_keeper_entity(self, proto_entity, keeper_entity):
+        # type: (proto.UserAlias, dict) -> None
+        _set_or_remove(keeper_entity, 'username', proto_entity.username)
+        _set_or_remove(keeper_entity, 'enterprise_user_id', proto_entity.enterpriseUserId)
+
+    def get_keeper_entity_id(self, entity):  # type: (dict) -> any
+        return entity['username']
+
+    def get_proto_entity_id(self, entity):  # type: (proto.UserAlias) -> any
+        return entity.username
+
+    def get_entity_type(self):
+        return proto.UserAlias
+
+    def get_keeper_entity_name(self):  # type: () -> str
+        return 'user_aliases'
+
 
 class _EnterpriseAdminApprovalRequestEntity(_EnterpriseEntity):
     def to_keeper_entity(self, proto_entity, keeper_entity):
