@@ -9,7 +9,6 @@
 # Contact: ops@keepersecurity.com
 #
 
-import collections
 import datetime
 import io
 import json
@@ -17,7 +16,7 @@ import logging
 import os.path
 import zipfile
 from contextlib import contextmanager
-from typing import Union, Optional, Tuple, Any, Dict, List, Set
+from typing import Union, Optional, Tuple, Any, Dict, Set, Iterable
 
 from ..importer import (
     BaseImporter, Record, Folder, RecordField, RecordReferences, SharedFolder, Attachment
@@ -34,6 +33,9 @@ class OnePasswordImporter(BaseImporter):
     def extension(self):
         return '1pux'
 
+    def support_folder_filter(self):
+        return True
+
     @staticmethod
     def get_field_value(field):  # type: (dict) -> Optional[Tuple[str, Any]]
         if isinstance(field, dict):
@@ -44,7 +46,7 @@ class OnePasswordImporter(BaseImporter):
                         return key, value[key]
 
     def do_import(self, filename, **kwargs):
-        # type: (BaseImporter, str, dict) -> collections.Iterable[Union[Record, SharedFolder]]
+        # type: (BaseImporter, str, dict) -> Iterable[Union[Record, SharedFolder]]
         record_types = {}
         params = kwargs.get('params')
         if isinstance(params, KeeperParams):
@@ -68,24 +70,27 @@ class OnePasswordImporter(BaseImporter):
                 return
             with zip_file.open('export.data', mode='r') as data_file:
                 export_data = json.load(data_file)
+
+            filter_folder = kwargs.get('filter_folder')
             if 'accounts' in export_data:
                 for account in export_data['accounts']:
-                    account_name = None      # type: Optional[str]
-                    if 'attrs' in account:
-                        account_name = account['attrs'].get('accountName')
                     for vault in account.get('vaults', []):
                         vault_name = None
                         is_shared = False
                         if 'attrs' in vault:
+                            name = vault['attrs'].get('name')
+                            if filter_folder and name != filter_folder:
+                                continue
+
                             vault_type = vault['attrs'].get('type', '')
                             if vault_type == 'E':
-                                vault_name = vault['attrs'].get('name')
+                                vault_name = name
                                 is_shared = True
                                 sf = SharedFolder()
                                 sf.path = vault_name
                                 yield sf
                             elif vault_type == 'U':
-                                vault_name = vault['attrs'].get('name')
+                                vault_name = name
 
                         vault_records = {}
                         references = {}   # type: Dict[str, Set[str]]
@@ -94,8 +99,11 @@ class OnePasswordImporter(BaseImporter):
                                 continue
                             record = Record()
                             record.uid = item.get('uuid') or utils.generate_uid()
+                            lm = item.get('updatedAt')
+                            if isinstance(lm, int) and lm > 0:
+                                record.last_modified = lm
                             category = item.get('categoryUuid', '')
-                            if category in ('001', '110', '112'):
+                            if category in ('001', '005', '110', '112'):
                                 record.type = 'login'
                             elif category == '002':
                                 record.type = 'bankCard'
@@ -105,6 +113,8 @@ class OnePasswordImporter(BaseImporter):
                                 record.type = 'contact'
                             elif category == '006':
                                 record.type = 'file'
+                            elif category == '100':
+                                record.type = 'softwareLicense'
                             elif category == '101':
                                 record.type = 'bankAccount'
                             elif category == '102':
@@ -183,6 +193,13 @@ class OnePasswordImporter(BaseImporter):
                                         record.login = value
                                     elif designation == 'password':
                                         record.password = value
+                            if 'password' in details:
+                                password = details['password']
+                                if password:
+                                    if record.password:
+                                        record.fields.append(RecordField(type='secret', label='Password', value=password))
+                                    else:
+                                        record.password = password
                             if 'notesPlain' in details:
                                 record.notes = details.get('notesPlain', '')
                             if 'documentAttributes' in details:
@@ -270,7 +287,7 @@ class OnePasswordImporter(BaseImporter):
                                                     if username:
                                                         record.fields.append(RecordField(type='login', label=f'{account_type} Username', value=username))
                                                     if password:
-                                                        record.fields.append(RecordField(type='password', label=f'{account_type} Password', value=password))
+                                                        record.fields.append(RecordField(type='secret', label=f'{account_type} Password', value=password))
                                                     if host:
                                                         record.fields.append(RecordField(type='host', label=f'{account_type} Server', value=host))
                                                 if security_value and security_value[1]:
@@ -367,7 +384,7 @@ class OnePasswordImporter(BaseImporter):
                                             elif field_id.endswith('website'):
                                                 ft = 'url'
                                                 fv = field_value
-                                            elif field_id in ('number', 'membership_no'):
+                                            elif field_id in ('number', 'membership_no', 'reg_code'):
                                                 if record.type in ('ssnCard', 'membership', 'driverLicense', 'healthInsurance', 'passport'):
                                                     ft = 'accountNumber'
                                                 elif record.type == 'softwareLicense':
@@ -406,7 +423,7 @@ class OnePasswordImporter(BaseImporter):
                                                         field_id == 'credential' or \
                                                         (field_id == 'pin' and record.type == 'membership'):
                                                     if record.password:
-                                                        ft = 'password'
+                                                        ft = 'secret'
                                                     else:
                                                         record.password = field_value
                                                         continue
@@ -461,6 +478,9 @@ class OnePasswordImporter(BaseImporter):
                                                     elif field_id == 'issue_date':
                                                         ft = 'date'
                                                         fl = 'dateIssued'
+                                                    elif field_id == 'order_date':
+                                                        ft = 'date'
+                                                        fl = 'dateActive'
                                                     else:
                                                         ft = 'date'
                                                 else:
@@ -534,37 +554,19 @@ class OnePasswordImporter(BaseImporter):
                             yield record
                         vault_records.clear()
 
-    @staticmethod
-    def adjust_field_label(record, field_type, field_label, fields):
-        # type: (Record, str, str, List[Dict]) -> str
-        if not isinstance(fields, list):
-            return field_label
-        if field_type == 'text':
-            return field_label
-        field = next((x for x in fields if x['$ref'] == field_type), None)
-        if not field:
-            return field_label
-        fl = field.get('label', '')
-        if fl == field_label:
-            return field_label
-        for f in record.fields:
-            if f.type == field_type and f.label == fl:
-                return field_label
-        return fl
-
 
 class OnePasswordAttachment(Attachment):
     def __init__(self, zip_file, zip_path, name, size):
         super().__init__()
         self.zip_file = zip_file
-        self.zip_path = zip_path
+        self.file_uid = zip_path
         self.name = name
         self.size = size
 
     @contextmanager
     def open(self):  # type: () -> io.BufferedIOBase
         with zipfile.ZipFile(self.zip_file, mode='r') as zip_file:
-            yield zip_file.open(self.zip_path, mode='r')
+            yield zip_file.open(self.file_uid, mode='r')
 
 
 
