@@ -9,6 +9,6 @@
 # Contact: ops@keepersecurity.com
 #
 
-from .commands import register_commands, register_command_info
+from .commands import register_commands, register_command_info, register_enterprise_commands
 
-__all__ = ['register_commands', 'register_command_info']
\ No newline at end of file
+__all__ = ['register_commands', 'register_command_info', 'register_enterprise_commands']
\ No newline at end of file
