@@ -10,15 +10,13 @@
 #
 
 import csv
-import json
 
 import sys
 
-from ...recordv3 import RecordV3
-
 from ..importer import (BaseFileImporter, BaseExporter, Record, Folder, RecordField, TWO_FACTOR_CODE,
                         FIELD_TYPE_ONE_TIME_CODE)
-
+from ...recordv3 import RecordV3
+from ... import record_types
 
 '''
 0 - folder
@@ -33,60 +31,6 @@ from ..importer import (BaseFileImporter, BaseExporter, Record, Folder, RecordFi
 
 
 class KeeperCsvImporter(BaseFileImporter):
-    @staticmethod
-    def import_field(field_type, field_value):  # type: (str, str) -> any
-        if not field_value:
-            return None
-        if not field_type:
-            return field_value
-        if field_type in {'text', 'multiline', 'secret'}:
-            return field_value
-        if field_type == 'date':
-            try:
-                return int(field_value)
-            except:
-                return None
-        if field_type == 'privateKey':
-            return KeeperCsvImporter.import_ssh_key_field(field_value)
-
-        str_values = field_value.split('\n')
-        values = []
-        for str_value in str_values:
-            if field_type == 'host':
-                v = KeeperCsvImporter.import_host_field(str_value)
-            elif field_type == 'phone':
-                v = KeeperCsvImporter.import_phone_field(str_value)
-            elif field_type == 'name':
-                v = KeeperCsvImporter.import_name_field(str_value)
-            elif field_type == 'address':
-                v = KeeperCsvImporter.import_address_field(str_value)
-            elif field_type == 'securityQuestion':
-                v = KeeperCsvImporter.import_q_and_a_field(str_value)
-            elif field_type == 'paymentCard':
-                v = KeeperCsvImporter.import_card_field(str_value)
-            elif field_type == 'bankAccount':
-                v = KeeperCsvImporter.import_account_field(str_value)
-            else:
-                v = str_value
-                if field_type in RecordV3.field_values:
-                    fv = RecordV3.field_values[field_type]
-                    if isinstance(fv.get('value'), dict):
-                        try:
-                            v = json.loads(str_value)
-                        except:
-                            pass
-                    elif isinstance(fv.get('value'), int):
-                        try:
-                            v = int(str_value)
-                        except:
-                            pass
-            if v:
-                values.append(v)
-        if values:
-            if len(values) == 1:
-                return values[0]
-            return values
-
     def do_import(self, filename, **kwargs):
         with open(filename, "r", encoding='utf-8-sig') as csvfile:
             reader = csv.reader(csvfile)
@@ -142,9 +86,9 @@ class KeeperCsvImporter(BaseFileImporter):
 
                                         field_type = ''
                                         if field.type:
-                                            t = RecordV3.get_field_type(field.type)
-                                            if isinstance(t, dict):
-                                                field_type = t.get('type', '')
+                                            t = record_types.RecordFields.get(field.type)
+                                            if t:
+                                                field_type = t.type
 
                                         field.value = KeeperCsvImporter.import_field(field_type, value)
                                         if field.value:
@@ -156,40 +100,7 @@ class KeeperCsvImporter(BaseFileImporter):
 
 
 class KeeperCsvExporter(BaseExporter):
-    @staticmethod
-    def export_field(field_type, field_value):  # type: (str, any) -> str
-        if not field_value:
-            return ''
-
-        if isinstance(field_value, str):
-            return field_value
-        if isinstance(field_value, list):
-            values = []
-            for value in field_value:
-                v = KeeperCsvExporter.export_field(field_type, value)
-                if v:
-                    values.append(v)
-            return '\n'.join((x.replace('\n', ' ') for x in values))
-        if isinstance(field_value, dict):
-            if field_type == 'host':
-                return BaseExporter.export_host_field(field_value)
-            if field_type == 'phone':
-                return BaseExporter.export_phone_field(field_value)
-            if field_type == 'name':
-                return BaseExporter.export_name_field(field_value)
-            if field_type == 'address':
-                return BaseExporter.export_address_field(field_value)
-            if field_type == 'securityQuestion':
-                return BaseExporter.export_q_and_a_field(field_value)
-            if field_type == 'paymentCard':
-                return BaseExporter.export_card_field(field_value)
-            if field_type == 'bankAccount':
-                return BaseExporter.export_account_field(field_value)
-            return json.dumps(field_value)
-
-        return str(field_value)
-
-    def do_export(self, filename, records, file_password=None):
+    def do_export(self, filename, records, **kwargs):
         csvfile = open(filename, 'w', encoding='utf-8', newline='') if filename else sys.stdout
         writer = csv.writer(csvfile)
         for r in records:
@@ -231,7 +142,7 @@ class KeeperCsvExporter(BaseExporter):
                             t = RecordV3.get_field_type(cf.type)
                             if isinstance(t, dict):
                                 field_type = t.get('type', '')
-                        field_value = KeeperCsvExporter.export_field(field_type, cf.value)
+                        field_value = BaseExporter.export_field(field_type, cf.value)
                         row.append(field_value)
                 writer.writerow(row)
         if filename:
