@@ -8,13 +8,15 @@
 # Copyright 2021 Keeper Security Inc.
 # Contact: ops@keepersecurity.com
 #
-
+import datetime
 import getpass
 import io
 import logging
 import os
 import re
 import uuid
+import urllib.parse
+from typing import Dict
 from xml.sax.saxutils import escape
 
 from pykeepass import PyKeePass
@@ -25,6 +27,8 @@ from pykeepass.group import Group
 from ..importer import path_components, PathDelimiter, BaseFileImporter, BaseExporter, \
     Record, Folder, SharedFolder, BytesAttachment, RecordField
 from ... import utils
+from ...display import bcolors
+from ...utils import parse_totp_uri
 
 _REFERENCE = r'\{REF:([TUPAN])@([IT]):([^\}]+)\}'
 
@@ -84,6 +88,10 @@ class KeepassImporter(BaseFileImporter):
                     if len(entries) > 0:
                         folder = KeepassImporter.get_folder(group)
                         for entry in entries:
+                            totp_secret = ''
+                            totp_issuer = ''
+                            totp_period = 0
+                            totp_digits = 0
                             record = Record()
                             fol = Folder()
                             fol.path = folder
@@ -100,22 +108,63 @@ class KeepassImporter(BaseFileImporter):
                                 record.login_url = entry.url
                             if entry.notes:
                                 record.notes = entry.notes
+                            if entry.otp:
+                                record.fields.append(RecordField('oneTimeCode', '', entry.otp))
+                            if hasattr(entry, 'ctime'):
+                                ctime = entry.ctime
+                                if isinstance(ctime, datetime.datetime):
+                                    record.last_modified = int(ctime.timestamp())
                             for key, value in entry.custom_properties.items():
+                                if key == '$type':
+                                    record.type = value
+                                    continue
+                                rest, sep, no = key.rpartition('#')
+                                if sep:
+                                    if no.isdigit():
+                                        key = rest
                                 if key.startswith('$'):
-                                    pos = key.find(':')
-                                    if pos > 0:
-                                        field_type = key[1:pos].strip()
-                                        field_label = key[pos+1:].strip()
-                                    else:
-                                        field_type = key[1:]
-                                        field_label = ''
+                                    ftype, sep, flabel = key.partition(':')
+                                    field_type = ftype[1:]
+                                    field_label = flabel
                                 else:
                                     field_type = ''
                                     field_label = key
+                                if field_label in ('TOTPSecret', 'TOTPPeriod', 'TOTPDigits', 'TOTPIssuer', 'ModifyTOTPSettings', 'ViewTOTPSettings'):
+                                    # Ignore TOTP custom fields (set via previous Keeper -> KDBX export implementation)
+                                    # if the entry's "otp" field contains the corresponding URI
+                                    if entry.otp:
+                                        continue
+                                    if field_label == 'TOTPSecret':
+                                        totp_secret = value
+                                    elif field_label == 'TOTPIssuer':
+                                        totp_issuer = value
+                                    elif field_label == 'TOTPPeriod':
+                                        try:
+                                            totp_period = int(value)
+                                        except:
+                                            pass
+                                    elif field_label == 'TOTPDigits':
+                                        try:
+                                            totp_digits = int(value)
+                                        except:
+                                            pass
+                                else:
+                                    field = RecordField()
+                                    field.type = field_type
+                                    field.label = field_label
+                                    field.value = KeepassImporter.import_field(field_type, value)
+                                    record.fields.append(field)
+                            if totp_secret:
+                                value = f'otpauth://totp/?secret={totp_secret}'
+                                if totp_issuer:
+                                    value += f'&issuer={urllib.parse.quote_plus(totp_issuer)}'
+                                if totp_period > 0:
+                                    value += f'&period={totp_period}'
+                                if totp_digits > 0:
+                                    value += f'&digits={totp_digits}'
                                 field = RecordField()
-                                field.type = field_type
-                                field.label = field_label
-                                field.value = value
+                                field.type = 'oneTimeCode'
+                                field.value = KeepassImporter.import_field(field_type, value)
                                 record.fields.append(field)
 
                             if entry.attachments:
@@ -123,7 +172,9 @@ class KeepassImporter(BaseFileImporter):
                                     if isinstance(a, KeepassAttachment):
                                         if record.attachments is None:
                                             record.attachments = []
-                                        record.attachments.append(BytesAttachment(a.filename, a.binary))
+                                        atta = BytesAttachment(a.filename, a.binary)
+                                        atta.file_uid = a.id
+                                        record.attachments.append(atta)
 
                             records.append(record)
                 id_map = None
@@ -214,11 +265,18 @@ class KeepassExporter(BaseExporter, XmlUtils):
         else:
             return str(keeper_value)
 
-    def do_export(self, filename, records, file_password=None):
+    def do_export(self, filename, records, file_password=None, **kwargs):
         master_password = file_password
-        if not master_password:
+        confirmed = True
+        while not master_password or not confirmed:
             print('Choose password for your Keepass file')
             master_password = getpass.getpass(prompt='...' + 'Keepass Password'.rjust(20) + ': ', stream=None)
+            print('\nRe-enter password for your Keepass file')
+            confirmation = getpass.getpass(prompt='...' + 'Keepass Password'.rjust(20) + ': ', stream=None)
+            confirmed = master_password == confirmation
+            retry_msg = 'The passwords you entered do not match.\nPlease try again.\n'
+            fail_msg = bcolors.FAIL + bcolors.BOLD + '\nALERT!\n' + retry_msg + bcolors.ENDC
+            not confirmed and print(fail_msg)
 
         sfs = []  # type: list[SharedFolder]
         rs = []   # type: list[Record]
@@ -257,20 +315,39 @@ class KeepassExporter(BaseExporter, XmlUtils):
                             break
 
                     if entry is None:
-                        entry = kdb.add_entry(node, r.title, r.login, r.password)
+                        entry = kdb.add_entry(node, title=r.title or '', username=r.login or '',
+                                              password=r.password or '', url=r.login_url or '',
+                                              notes=r.notes or '')
                         if r.uid:
                             entry.UUID = uuid.UUID(bytes=utils.base64_url_decode(r.uid))
-                    entry.url = r.login_url
-                    entry.notes = r.notes
+                    if r.type:
+                        entry.set_custom_property('$type', r.type)
                     if r.fields:
+                        custom_names = {}   # type: Dict[str, int]
                         for cf in r.fields:
+                            if cf.type == 'oneTimeCode':
+                                entry.otp = cf.value
+                                # Set custom fields for Pleasant Password TOTP compatibility
+                                totp_props = parse_totp_uri(cf.value)
+                                for key in ['secret', 'period', 'issuer', 'digits']:
+                                    val = totp_props.get(key)
+                                    val and entry.set_custom_property(f'TOTP{key.capitalize()}', str(val))
+
+                                continue
                             if cf.type and cf.label:
                                 title = f'${cf.type}:{cf.label}'
                             elif cf.type:
                                 title = f'${cf.type}'
                             else:
                                 title = cf.label or ''
-                            entry.set_custom_property(title, self.to_keepass_value(cf.value))
+                            if title in custom_names:
+                                no = custom_names[title]
+                                no += 1
+                                custom_names[title] = no
+                                title = f'{title}#{no}'
+                            else:
+                                custom_names[title] = 1
+                            entry.set_custom_property(title, self.to_keepass_value(BaseExporter.export_field(cf.type, cf.value)))
 
                     if r.attachments:
                         for atta in r.attachments:
