@@ -13,4 +13,14 @@ class Account(object):
         self.attach_key = attach_key
         self.totp_secret = totp_secret
         self.totp_url = totp_url
+        self.last_modified = 0
         self.attachments = []
+        self.custom_fields = []
+
+
+class CustomField(object):
+    def __init__(self, f_name, f_type, f_value, checked):
+        self.name = f_name
+        self.type = f_type
+        self.value = f_value
+        self.checked = checked
