@@ -17,13 +17,14 @@ import tempfile
 from glob import glob
 from typing import Optional, List
 
+from ...params import KeeperParams
 from ..importer import (
-    BaseImporter, Record, Folder, RecordField, RecordReferences, SharedFolder, Permission, replace_email_domain,
-    FIELD_TYPE_ONE_TIME_CODE
-)
+    BaseImporter, Record, Folder, RecordField, RecordReferences, SharedFolder, Permission, BaseDownloadMembership,
+    replace_email_domain, FIELD_TYPE_ONE_TIME_CODE)
 from .account import Account
 from .exceptions import LastPassUnknownError
 from .vault import Vault, TMPDIR_PREFIX
+from . import fetcher
 
 
 class LastPassImporter(BaseImporter):
@@ -131,6 +132,15 @@ class LastPassImporter(BaseImporter):
             logging.warning(warn_msg)
 
     def do_import(self, name, users_only=False, old_domain=None, new_domain=None, tmpdir=None, **kwargs):
+        dry_run = kwargs.get('dry_run') or False
+        request_settings = {}
+        params = kwargs.get('params')
+        if isinstance(params, KeeperParams):
+            request_settings['proxies'] = params.rest_context.proxies
+            request_settings['certificate_check'] = params.rest_context.certificate_check
+        if 'filter_folder' in kwargs and kwargs['filter_folder']:
+            request_settings['filter_folder'] = kwargs['filter_folder']
+
         self.tmpdir = tmpdir
         username = name
         password = getpass.getpass(prompt='...' + 'LastPass Password'.rjust(30) + ': ', stream=None)
@@ -140,11 +150,14 @@ class LastPassImporter(BaseImporter):
             twofa_code = None
 
         try:
-            vault = Vault.open_remote(
-                username, password, multifactor_password=twofa_code, users_only=users_only, tmpdir=tmpdir
-            )
+            vault = Vault.open_remote(username, password, multifactor_password=twofa_code, users_only=users_only,
+                                      tmpdir=tmpdir, get_attachments=not dry_run, **request_settings)
         except LastPassUnknownError as lpe:
-            logging.warning(lpe)
+            message = str(lpe)
+            if message.startswith('Try again OR look for an email'):
+                message += 'If you do not receive an email, go to LastPass > Account Settings > Advanced Settings ' \
+                           'and ensure that "Disable Email Verification" is unchecked.'
+            logging.warning(message)
             return
         else:
             self.vault = vault
@@ -153,60 +166,83 @@ class LastPassImporter(BaseImporter):
                 logging.warning(f'The following errors occurred retrieving Lastpass shared folder members:\n{err_list}')
 
         for shared_folder in vault.shared_folders:
-            folder = SharedFolder()
-            folder.path = shared_folder.name
-            folder.permissions = []
-            if shared_folder.members:
-                for member in shared_folder.members:
-                    perm = Permission()
-                    perm.name = replace_email_domain(member['username'], old_domain, new_domain)
-                    perm.manage_records = member['readonly'] == '0'
-                    perm.manage_users = member['can_administer'] == '1'
-                    folder.permissions.append(perm)
-            if shared_folder.teams:
-                for team in shared_folder.teams:
-                    perm = Permission()
-                    perm.name = team['name']
-                    perm.manage_records = team['readonly'] == '0'
-                    perm.manage_users = team['can_administer'] == '1'
-                    folder.permissions.append(perm)
+            if shared_folder.name:
+                folder = SharedFolder()
+                folder.path = shared_folder.name
 
-            yield folder
+                folder.permissions = []
+                if shared_folder.members:
+                    for member in shared_folder.members:
+                        perm = Permission()
+                        perm.name = replace_email_domain(member['username'], old_domain, new_domain)
+                        perm.manage_records = member['readonly'] == '0'
+                        perm.manage_users = member['can_administer'] == '1'
+                        folder.permissions.append(perm)
+                if shared_folder.teams:
+                    for team in shared_folder.teams:
+                        perm = Permission()
+                        perm.name = team['name']
+                        perm.manage_records = team['readonly'] == '0'
+                        perm.manage_users = team['can_administer'] == '1'
+                        folder.permissions.append(perm)
+
+                yield folder
 
         missing_titles = 0
         for account in vault.accounts:  # type: Account
             record = Record()
             is_secure_note = False
             if account.url:
-                record.login_url = account.url.decode('utf-8')
+                record.login_url = account.url.decode('utf-8', 'ignore')
                 if record.login_url == 'http://sn':
                     is_secure_note = True
                     record.login_url = None
                 elif record.login_url == 'http://group':
                     continue
 
+            record.type = 'login'
             if account.id:
                 record.uid = account.id
             if account.name:
-                record.title = account.name.decode('utf-8')
+                record.title = account.name.decode('utf-8', 'ignore')
             else:
                 missing_titles += 1
                 record.title = f'Missing Title {missing_titles}'
                 logging.warning(f'Missing title in record from LastPass. Assigning title "{record.title}"')
             if account.username:
-                record.login = account.username.decode('utf-8')
+                record.login = account.username.decode('utf-8', 'ignore')
             if account.password:
-                record.password = account.password.decode('utf-8')
+                record.password = account.password.decode('utf-8', 'ignore')
             if account.totp_url:
                 record.fields.append(
                     RecordField(type=FIELD_TYPE_ONE_TIME_CODE, value=account.totp_url)
                 )
+            if isinstance(account.last_modified, int) and account.last_modified > 0:
+                record.last_modified = account.last_modified
+            if isinstance(account.custom_fields, list):
+                for cf in account.custom_fields:
+                    field_label = cf.name
+                    if cf.type == 'password':
+                        field_type = 'secret'
+                    elif cf.type == 'email':
+                        field_type = 'email'
+                    elif cf.type == 'textarea':
+                        field_type = 'multiline'
+                    elif cf.type == 'tel':
+                        field_type = 'phone'
+                    else:
+                        field_type = 'text'
+                    cf = RecordField(type=field_type, label=field_label, value=cf.value)
+                    record.fields.append(cf)
             if len(account.attachments) > 0:
                 if record.attachments is None:
                     record.attachments = []
                 record.attachments = account.attachments
             if account.notes:
-                notes = account.notes.decode('utf-8')
+                try:
+                    notes = account.notes.decode('utf-8', 'ignore')
+                except UnicodeDecodeError:
+                    notes = ''
                 if notes.startswith('NoteType:'):
                     typed_values = self.parse_typed_notes(notes)
                     if 'NoteType' in typed_values:
@@ -246,8 +282,6 @@ class LastPassImporter(BaseImporter):
                             self.populate_ssh_key(record, typed_values)
                         elif note_type == 'Software License':
                             self.populate_software_license(record, typed_values)
-                        else:
-                            record.type = 'login'
 
                     username = typed_values.pop('Username', '')
                     if username:
@@ -307,7 +341,9 @@ class LastPassImporter(BaseImporter):
                 if account.shared_folder:
                     fol.domain = account.shared_folder.name
                 if account.group:
-                    fol.path = account.group.decode('utf-8')
+                    fol.path = account.group
+                    if isinstance(fol.path, bytes):
+                        fol.path = fol.path.decode('utf-8', 'ignore')
                 record.folders = [fol]
 
             yield record
@@ -400,9 +436,6 @@ class LastPassImporter(BaseImporter):
             'accountNumber': notes.pop('Account Number', ''),
         }
         record.fields.append(bank)
-        bank_name = notes.pop('Bank Name', '')
-        if bank_name:
-            record.title = bank_name
 
     def populate_passport(self, record, notes): # type: (Record, dict) -> None
         record.type = 'passport'
@@ -523,7 +556,7 @@ class LastPassImporter(BaseImporter):
         record.fields.append(host)
         record.login_url = ''
 
-    def populate_server(self, record, notes):  # type: (str, Record, dict) -> None
+    def populate_server(self, record, notes):  # type: (Record, dict) -> None
         record.type = 'serverCredentials'
         host = RecordField(type='host')
         host.value = {
@@ -638,6 +671,11 @@ class LastPassAddress(object):
         if 'Address 1' in notes:
             address.street1 = notes.pop('Address 1', '')
             address.street2 = notes.pop('Address 2', '')
+            street3 = notes.pop('Address 3', '')
+            if street3:
+                address.street2 += f' {street3}'
+                address.street2 = address.street2.strip()
+
         elif 'Address' in notes:
             s1, sep, s2 = notes.pop('Address', '').partition(',')
             address.street1 = s1.strip()
@@ -652,3 +690,65 @@ class LastPassAddress(object):
         address.country = notes.pop('Country', '')
 
         return address
+
+
+class LastpassMembershipDownload(BaseDownloadMembership):
+    def download_membership(self, params, **kwargs):
+        username = input('...' + 'LastPass Username'.rjust(30) + ': ')
+        if not username:
+            logging.warning('LastPass username is required')
+            return
+        password = getpass.getpass(prompt='...' + 'LastPass Password'.rjust(30) + ': ', stream=None)
+        if not password:
+            logging.warning('LastPass password is required')
+            return
+
+        print('Press <Enter> if account is not protected with Multifactor Authentication')
+        twofa_code = getpass.getpass(prompt='...' + 'Multifactor Password'.rjust(30) + ': ', stream=None)
+        if not twofa_code:
+            twofa_code = None
+
+        session = None
+        request_settings = {
+            'proxies': params.rest_context.proxies,
+            'certificate_check': params.rest_context.certificate_check
+        }
+        try:
+            session = fetcher.login(username, password, twofa_code, **request_settings)
+            blob = fetcher.fetch(session, **request_settings)
+            encryption_key = blob.encryption_key(username, password)
+            vault = Vault(blob, encryption_key, session, shared_folder_details=False, get_attachments=False)
+
+            lastpass_shared_folder = [x for x in vault.shared_folders]
+
+            for lpsf in lastpass_shared_folder:
+                logging.info('Loading shared folder membership for "%s"', lpsf.name)
+
+                members, teams, error = fetcher.fetch_shared_folder_members(session, lpsf.id, **request_settings)
+                sf = SharedFolder()
+                sf.uid = lpsf.id
+                sf.path = lpsf.name
+                sf.permissions = []
+                if members:
+                    sf.permissions.extend((self._lastpass_permission(x) for x in members))
+                if teams:
+                    sf.permissions.extend((self._lastpass_permission(x, team=True) for x in teams))
+                yield sf
+
+        except Exception as e:
+            logging.warning(e)
+        finally:
+            if session:
+                fetcher.logout(session, **request_settings)
+
+    @staticmethod
+    def _lastpass_permission(lp_permission, team=False):  # type: (dict, Optional[bool]) -> Permission
+        permission = Permission()
+        if team:
+            permission.name = lp_permission['name']
+        else:
+            permission.name = lp_permission['username']
+        permission.manage_records = lp_permission['readonly'] == '0'
+        permission.manage_users = lp_permission['can_administer'] == '1'
+        return permission
+
