@@ -1,21 +1,26 @@
 # coding: utf-8
-from base64 import b64decode
 import binascii
 import codecs
-from io import BytesIO
-import struct
+import logging
 import re
+import struct
+from base64 import b64decode
+from io import BytesIO
+from typing import Any, Optional, Tuple
 from urllib.parse import urlunsplit, urlencode
 
-from Cryptodome.Cipher import AES, PKCS1_OAEP
-from Cryptodome.Util import number
-from Cryptodome.PublicKey import RSA
+from cryptography.hazmat.backends import default_backend
+from cryptography.hazmat.primitives import serialization
+from cryptography.hazmat.primitives.asymmetric.padding import OAEP, MGF1
+from cryptography.hazmat.primitives.ciphers import Cipher
+from cryptography.hazmat.primitives.ciphers.algorithms import AES
+from cryptography.hazmat.primitives.ciphers.modes import CBC, ECB
+from cryptography.hazmat.primitives.hashes import SHA1
 
-from .account import Account
+from .account import Account, CustomField
 from .attachment import LastpassAttachment
 from .chunk import Chunk
 
-
 # Secure note types that contain account-like information
 ALLOWED_SECURE_NOTE_TYPES = [
     b"Server",
@@ -61,14 +66,25 @@ def parse_ACCT(chunk, encryption_key, shared_folder):
     io = BytesIO(chunk.payload)
     id = read_item(io).decode('utf-8')
     name = decode_aes256_plain_auto(read_item(io), encryption_key)
-    group = decode_aes256_plain_auto(read_item(io), encryption_key)
-    url = decode_hex(read_item(io))
+    try:
+        group = decode_aes256_plain_auto(read_item(io), encryption_key).decode('utf-8', 'ignore')
+    except:
+        pass
+    url_enc = read_item(io)
+    if url_enc:
+        if url_enc[0] == ord('!'):
+            url = decode_aes256_plain_auto(url_enc, encryption_key)
+        else:
+            url = decode_hex(url_enc)
     notes = decode_aes256_plain_auto(read_item(io), encryption_key)
+    # fav, ?
     skip_item(io, 2)
     username = decode_aes256_plain_auto(read_item(io), encryption_key)
     password = decode_aes256_plain_auto(read_item(io), encryption_key)
+    # pwprotect, ?
     skip_item(io, 2)
     secure_note = read_item(io)
+    # last_touch, ...
     skip_item(io, 14)
 
     attach_key_encrypted = read_item(io)
@@ -77,8 +93,15 @@ def parse_ACCT(chunk, encryption_key, shared_folder):
         attach_key = decode_hex(decode_aes256_base64_auto(attach_key_encrypted, encryption_key))
     else:
         attach_key = None
-
-    skip_item(io, 11)
+    skip_item(io, 3)
+    last_modified = 0
+    try:
+        lm = read_item(io).decode('utf-8')    # type: str
+        if lm.isnumeric():
+            last_modified = int(lm)
+    except:
+        pass
+    skip_item(io, 7)
     try:
         totp_secret = decode_aes256_plain_auto(read_item(io), encryption_key).decode('utf-8')
     except Exception:
@@ -100,7 +123,10 @@ def parse_ACCT(chunk, encryption_key, shared_folder):
             username = parsed.get('username', username)
             password = parsed.get('password', password)
 
-    return Account(id, name, username, password, url, group, notes, shared_folder, attach_key, totp_secret, totp_url)
+    account = Account(id, name, username, password, url, group, notes, shared_folder, attach_key, totp_secret, totp_url)
+    if isinstance(last_modified, int) and last_modified > 0:
+        account.last_modified = last_modified
+    return account
 
 
 def parse_PRIK(chunk, encryption_key):
@@ -111,16 +137,10 @@ def parse_PRIK(chunk, encryption_key):
                               encryption_key)
 
     hex_key = re.match(br'^LastPassPrivateKey<(?P<hex_key>.*)>LastPassPrivateKey$', decrypted).group('hex_key')
-    rsa_key = RSA.importKey(decode_hex(hex_key))
-
-    rsa_key.dmp1 = rsa_key.d % (rsa_key.p - 1)
-    rsa_key.dmq1 = rsa_key.d % (rsa_key.q - 1)
-    rsa_key.iqmp = number.inverse(rsa_key.q, rsa_key.p)
-
-    return rsa_key
+    return decode_hex(hex_key)
 
 
-def parse_SHAR(chunk, encryption_key, rsa_key):
+def parse_SHAR(chunk, encryption_key, rsa_key):   # type: (Any, bytes, bytes) -> Any
     # TODO: Fake some data and make a test
     io = BytesIO(chunk.payload)
     id = read_item(io)
@@ -133,38 +153,60 @@ def parse_SHAR(chunk, encryption_key, rsa_key):
     # where it's only AES encrypted with the regular encryption key.
     # When the key is blank, then there's a RSA encrypted key, which has to
     # be decrypted first before use.
-    if not key:
-        key = decode_hex(PKCS1_OAEP.new(rsa_key).decrypt(encrypted_key))
-    else:
-        key = decode_hex(decode_aes256_plain_auto(key, encryption_key))
+    try:
+        if key:
+            try:
+                key = decode_hex(decode_aes256_plain_auto(key, encryption_key))
+            except:
+                key = ''
+        if not key:
+            key = decode_hex(decode_rsa_plain_oaep(encrypted_key, rsa_key))
 
-    name = decode_aes256_base64_auto(encrypted_name, key)
-
-    # TODO: Return an object, not a dict
-    return {'id': id, 'name': name, 'encryption_key': key}
+        name = decode_aes256_base64_auto(encrypted_name, key)
 
+        # TODO: Return an object, not a dict
+        return {'id': id, 'name': name, 'encryption_key': key}
+    except Exception as e:
+        logging.warning('Shared folder decryption error: %s', e)
 
 def parse_ATTA(chunk, accounts):
     attachment = None
     io = BytesIO(chunk.payload)
-    id = read_item(io).decode('utf-8')
-    parent_id = read_item(io).decode('utf-8')
-    mimetype = read_item(io).decode('utf-8')
-    storagekey = read_item(io).decode('utf-8')
-    lastpass_size = int(read_item(io))
+    id = read_item(io).decode('utf-8', 'ignore')
+    parent_id = read_item(io).decode('utf-8', 'ignore')
+    mimetype = read_item(io).decode('utf-8', 'ignore')
+    storagekey = read_item(io).decode('utf-8', 'ignore')
+    size = read_item(io)
+    try:
+        lastpass_size = int(size)
+    except ValueError:
+        return None
     filename_encrypted = read_item(io)
 
     parents = [a for a in accounts if a.id == parent_id]
     if len(parents) > 0:
         parent = parents[0]
         if parent.attach_key:
-            filename = decode_aes256_base64_auto(filename_encrypted, parent.attach_key).decode('utf-8')
+            filename = decode_aes256_base64_auto(filename_encrypted, parent.attach_key).decode('utf-8', 'ignore')
             attachment = LastpassAttachment(id, parent, mimetype, storagekey, lastpass_size, filename)
             parent.attachments.append(attachment)
 
     return attachment
 
 
+def parse_ACFL(chunk, encryption_key):
+    io = BytesIO(chunk.payload)
+    field_name = read_item(io).decode('utf-8', 'ignore')
+    field_type = read_item(io).decode('utf-8', 'ignore')
+    if field_type in {'email', 'tel', 'text', 'password', 'textarea'}:
+        field_value = decode_aes256_plain_auto(read_item(io), encryption_key)
+    else:
+        field_value = read_item(io)
+    field_value = field_value.decode('utf-8', 'ignore')
+    checked = read_item(io).decode('utf-8', 'ignore')
+    return CustomField(field_name, field_type, field_value, checked == '1')
+
+
 def parse_secure_note_server(notes):
     info = {}
 
@@ -327,12 +369,19 @@ def decode_aes256(cipher, iv, data, encryption_key):
     If for :ecb iv is not used and should be set to "".
     """
     if cipher == 'cbc':
-        aes = AES.new(encryption_key, AES.MODE_CBC, iv)
+        aes_cipher = Cipher(AES(encryption_key), CBC(iv), backend=default_backend())
     elif cipher == 'ecb':
-        aes = AES.new(encryption_key, AES.MODE_ECB)
+        aes_cipher = Cipher(AES(encryption_key), ECB(), backend=default_backend())
     else:
         raise ValueError('Unknown AES mode')
-    d = aes.decrypt(data)
+    decrypter = aes_cipher.decryptor()
+    d = decrypter.update(data) + decrypter.finalize()
     # http://passingcuriosity.com/2009/aes-encryption-in-python-with-m2crypto/
     unpad = lambda s: s[0:-ord(d[-1:])]
     return unpad(d)
+
+
+def decode_rsa_plain_oaep(data, der_private_key):
+    assert isinstance(data, bytes)
+    rsa_key = serialization.load_der_private_key(der_private_key, None, default_backend())
+    return rsa_key.decrypt(data, OAEP(mgf=MGF1(algorithm=SHA1()), algorithm=SHA1(), label=None))
