@@ -3,7 +3,9 @@ import json
 import logging
 import os
 import shutil
+import sys
 from tempfile import mkdtemp
+from typing import Optional
 
 from . import fetcher
 from . import parser
@@ -19,15 +21,12 @@ class Vault(object):
     @classmethod
     def open_remote(cls, username, password, multifactor_password=None, client_id=None, **kwargs):
         """Fetches a blob from the server and creates a vault"""
-        session = fetcher.login(username, password, multifactor_password, client_id)
-        blob = fetcher.fetch(session)
+        session = fetcher.login(username, password, multifactor_password, client_id, **kwargs)
+        blob = fetcher.fetch(session, **kwargs)
         encryption_key = blob.encryption_key(username, password)
-        vault = cls(
-            blob, encryption_key, session, tmpdir=kwargs.get('tmpdir'),
-            shared_folder_details=kwargs.get('users_only') or False
-        )
+        vault = cls(blob, encryption_key, session, **kwargs)
 
-        fetcher.logout(session)
+        fetcher.logout(session, **kwargs)
         return vault
 
     @classmethod
@@ -36,7 +35,8 @@ class Vault(object):
         # TODO: read the blob here
         raise NotImplementedError()
 
-    def __init__(self, blob, encryption_key, session, tmpdir=None, shared_folder_details=False, get_attachments=True):
+    def __init__(self, blob, encryption_key, session, tmpdir=None, shared_folder_details=False, get_attachments=True,
+                 **kwargs):
         """This more of an internal method, use one of the static constructors instead"""
         chunks = parser.extract_chunks(blob)
 
@@ -48,14 +48,29 @@ class Vault(object):
         self.attachments = []
         self.accounts = self.parse_accounts(chunks, encryption_key)
         self.tmpdir = None
+        self.proxies = kwargs.get('proxies')
+        self.certificate_check = kwargs.get('certificate_check')
+
+        if 'filter_folder' in kwargs and kwargs['filter_folder']:
+            folder_name = kwargs['filter_folder'].lower()
+            shared_folder = next((x for x in self.shared_folders if x.name.lower() == folder_name), None)
+            if isinstance(shared_folder, LastpassSharedFolder):
+                self.shared_folders = [shared_folder]
+                self.accounts = [x for x in self.accounts if x.shared_folder is shared_folder]
+            else:
+                self.shared_folders = []
+                self.accounts = [x for x in self.accounts if not x.shared_folder and x.group and x.group.lower() == folder_name]
+            account_ids = {x.id for x in self.accounts}
+            self.attachments = [x for x in self.attachments if x.parent.id in account_ids]
 
         if get_attachments:
-            self.process_attachments(session, tmpdir)
+            self.process_attachments(session, tmpdir, proxies=self.proxies, certificate_check=self.certificate_check)
 
         try:
             if self.shared_folders and shared_folder_details:
                 for shared_folder in self.shared_folders:
-                    members, teams, error = fetcher.fetch_shared_folder_members(session, shared_folder.id)
+                    members, teams, error = fetcher.fetch_shared_folder_members(
+                        session, shared_folder.id, proxies=self.proxies, certificate_check=self.certificate_check)
                     if error:
                         self.errors.add(error)
                         break
@@ -72,27 +87,47 @@ class Vault(object):
         accounts = []
 
         key = encryption_key
-        rsa_private_key = None
+        rsa_private_key = None   # type: Optional[bytes]
         shared_folder = None
-
+        last_account = None
         for i in chunks:
             if i.id == b'ACCT':
-                account = parser.parse_ACCT(i, key, shared_folder)
-                if account:
-                    accounts.append(account)
+                try:
+                    last_account = parser.parse_ACCT(i, key, shared_folder)
+                except Exception as e:
+                    logging.debug('Account parse error: %s', e)
+                    last_account = None
+                if last_account:
+                    accounts.append(last_account)
             elif i.id == b'PRIK':
                 rsa_private_key = parser.parse_PRIK(i, encryption_key)
             elif i.id == b'SHAR':
-                # After SHAR chunk all the folliwing accounts are enrypted with a new key
+                # After SHAR chunk all the following accounts are encrypted with a new key
                 share = parser.parse_SHAR(i, encryption_key, rsa_private_key)
-                key = share['encryption_key']
-                shareid = share['id'].decode('utf-8')
-                shared_folder = LastpassSharedFolder(shareid, share['name'].decode('utf-8'))
-                self.shared_folders.append(shared_folder)
+                if share:
+                    key = share['encryption_key']
+                    shareid = share['id'].decode('utf-8')
+                    share_name = share['name'].decode('utf-8', 'ignore')
+                    share_name = share_name.strip()
+                    shared_folder = LastpassSharedFolder(shareid, share_name)
+                    self.shared_folders.append(shared_folder)
+                else:
+                    shared_folder = None
+                    key = encryption_key
             elif i.id == b'ATTA':
                 attachment = parser.parse_ATTA(i, accounts)
                 if attachment:
                     self.attachments.append(attachment)
+            elif i.id in (b'ACFL', b'ACOF'):
+                if last_account:
+                    try:
+                        cf = parser.parse_ACFL(i, key)
+                        if cf:
+                            last_account.custom_fields.append(cf)
+                    except Exception as e:
+                        logging.debug('Error parsing custom field ID: %s: %s', i.id.decode(), e)
+            else:
+                pass
 
         return accounts
 
@@ -101,9 +136,12 @@ class Vault(object):
         if self.tmpdir:
             shutil.rmtree(self.tmpdir, ignore_errors=True)
 
-    def process_attachments(self, session, tmpdir=None):
+    def process_attachments(self, session, tmpdir=None, **kwargs):
         skip_bad_attachments = []
         attach_cnt = len(self.attachments)
+        if attach_cnt == 0:
+            return
+
         if tmpdir is None:
             self.tmpdir = mkdtemp(prefix=TMPDIR_PREFIX)
         else:
@@ -127,12 +165,12 @@ class Vault(object):
             if os.path.isfile(attachment.tmpfile) and os.path.getsize(attachment.tmpfile) == attachment.lastpass_size:
                 print(f'{i + 1}. Found {attachment.name}')
             else:
-                attachment_stream = fetcher.stream_attachment(session, attachment)
+                attachment_stream = fetcher.stream_attachment(session, attachment, **kwargs)
                 if attachment_stream:
                     try:
                         with attachment_stream as r:
                             with open(attachment.tmpfile, 'wb') as f:
-                                print(f'{i + 1}. Downloading {attachment.name} ... ', end='', flush=True)
+                                print(f'{i + 1}. Downloading {attachment.name} ... ', file=sys.stderr, end='', flush=True)
                                 shutil.copyfileobj(r.raw, f)
                                 print('Done')
                     except Exception as e:
