@@ -34,10 +34,22 @@ class AWSRotator:
         """Set session and iam properties for instance"""
         try:
             self.session = set_session(self.aws_profile)
-            self.iam = boto3.client('iam')
         except Exception as e:
             logging.error(f'Unable to create AWS session using {self.profile_msg}: {e}')
             return False
         else:
             logging.debug(f'Created AWS session using {self.profile_msg}')
             return True
+
+    def assume_role(self, role_arn, session_name="RotationSession"):
+        sts_client = self.session.client('sts')
+        assumed_role_object = sts_client.assume_role(
+            RoleArn=role_arn,
+            RoleSessionName=session_name
+        )
+        credentials = assumed_role_object['Credentials']
+        self.session = boto3.Session(
+            aws_access_key_id=credentials['AccessKeyId'],
+            aws_secret_access_key=credentials['SecretAccessKey'],
+            aws_session_token=credentials['SessionToken']
+        )
