@@ -12,16 +12,14 @@ import datetime
 import hashlib
 import hmac
 import itertools
+from typing import Tuple, Optional
 
 from urllib import parse
 
 from .base32hex import b32decode
-from .error import Error
-from .subfolder import get_folder_path, find_folders, BaseFolderNode
 
 
-def get_totp_code(url):
-    # type: (str) -> (str, int) or None
+def get_totp_code(url, offset=None):   # type: (str, Optional[int]) -> Optional[Tuple[str, int, int]]
     comp = parse.urlparse(url)
     if comp.scheme == 'otpauth':
         secret = None
@@ -40,6 +38,8 @@ def get_totp_code(url):
         if secret:
             tm_base = int(datetime.datetime.now().timestamp())
             tm = tm_base / period
+            if isinstance(offset, int):
+                tm += offset
             alg = algorithm.lower()
             if alg in hashlib.__dict__:
                 reminder = len(secret) % 8
@@ -60,7 +60,7 @@ def get_totp_code(url):
                     code = code.rjust(digits, '0')
                 return code, period - (tm_base % period), period
             else:
-                raise Error('Unsupported hash algorithm: {0}'.format(algorithm))
+                raise Exception(f'Unsupported hash algorithm: {algorithm}')
 
 
 class Record:
@@ -197,20 +197,13 @@ class Record:
             if len(idxs) == 1:
                 return self.custom_fields.pop(idxs[0])
 
-    def display(self, **kwargs):
+    def display(self, unmask=False):
         print('')
         print('{0:>20s}: {1:<20s}'.format('UID', self.record_uid))
-        params = None
-        if 'params' in kwargs:
-            params = kwargs['params']
-            folders = [get_folder_path(params, x) for x in find_folders(params, self.record_uid)]
-            for i in range(len(folders)):
-                print('{0:>21s} {1:<20s}'.format('Folder:' if i == 0 else '', folders[i]))
-
         print('{0:>20s}: {1:<20s}'.format('Type', ''))
         if self.title: print('{0:>20s}: {1:<20s}'.format('Title', self.title))
         if self.login: print('{0:>20s}: {1:<20s}'.format('Login', self.login))
-        if self.password: print('{0:>20s}: {1:<20s}'.format('Password', self.password if kwargs.get('unmask') else '********'))
+        if self.password: print('{0:>20s}: {1:<20s}'.format('Password', self.password if unmask else '********'))
         if self.login_url: print('{0:>20s}: {1:<20s}'.format('URL', self.login_url))
         # print('{0:>20s}: https://keepersecurity.com/vault#detail/{1}'.format('Link',self.record_uid))
 
@@ -241,83 +234,15 @@ class Record:
                         size = size / 1024
                         scale = 'Gb'
                 sz = '{0:.2f}'.format(size).rstrip('0').rstrip('.')
-                print('{0:>21s} {1:<20s} {2:>6s}{3:<2s} {4:>6s}: {5}'.format('Attachments:' if i == 0 else '',
-                                                                             atta.get('title') or atta.get('name'),
-                                                                             sz, scale, 'ID',
-                                                                             atta.get('id')))
+                print('{0:>21s} {1:<20s} {2:>6s}{3:<2s} {4:>6s}: {5}'.format(
+                    'Attachments:' if i == 0 else '', atta.get('title') or atta.get('name'), sz, scale, 'ID', atta.get('id')))
 
         if self.totp:
-            print('{0:>20s}: {1}'.format('TOTP URL', self.totp))
+            print('{0:>20s}: {1}'.format('TOTP URL', self.totp if unmask else '********'))
             code, remain, _ = get_totp_code(self.totp)
             if code:
                 print('{0:>20s}: {1:<20s} valid for {2} sec'.format('Two Factor Code', code, remain))
 
-        if params is not None:
-            if self.record_uid in params.record_cache:
-                rec = params.record_cache[self.record_uid]
-                if 'shares' in rec:
-                    no = 0
-                    if 'user_permissions' in rec['shares']:
-                        perm = rec['shares']['user_permissions'].copy()
-                        perm.sort(key=lambda r: (' 1' if r.get('owner') else ' 2' if r.get(
-                            'editable') else ' 3' if r.get('sharable') else '') + r.get('username'))
-                        for uo in perm:
-                            flags = ''
-                            if uo.get('owner'):
-                                flags = 'Owner'
-                            elif uo.get('awaiting_approval'):
-                                flags = 'Awaiting Approval'
-                            else:
-                                if uo.get('editable'):
-                                    flags = 'Edit'
-                                if uo.get('sharable'):
-                                    if flags:
-                                        flags = flags + ', '
-                                    flags = flags + 'Share'
-                            if not flags:
-                                flags = 'View'
-
-                            print('{0:>21s} {1} ({2}) {3}'.format('Shared Users:' if no == 0 else '', uo['username'],
-                                                                  flags,
-                                                                  'self' if uo['username'] == params.user else ''))
-                            no = no + 1
-                    no = 0
-                    if 'shared_folder_permissions' in rec['shares']:
-                        for sfo in rec['shares']['shared_folder_permissions']:
-                            flags = ''
-                            if sfo.get('editable'):
-                                flags = 'Edit'
-                            if sfo.get('reshareable'):
-                                if flags:
-                                    flags = flags + ', '
-                                flags = flags + 'Share'
-                            if not flags:
-                                flags = 'View'
-                            sf_uid = sfo['shared_folder_uid']
-                            for f_uid in find_folders(params, self.record_uid):
-                                if f_uid in params.subfolder_cache:
-                                    fol = params.folder_cache[f_uid]
-                                    if fol.type in {BaseFolderNode.SharedFolderType,
-                                                    BaseFolderNode.SharedFolderFolderType}:
-                                        sfid = fol.uid if fol.type == BaseFolderNode.SharedFolderType else fol.shared_folder_uid
-                                        if sf_uid == sfid:
-                                            print('{0:>21s} {1:<20s}'.format('Shared Folders:' if no == 0 else '',
-                                                                             fol.name))
-                                            no = no + 1
-
-        if params and params.breach_watch:
-            bw_status = params.breach_watch.get_record_status(params, self.record_uid)
-            if bw_status and 'status' in bw_status:
-                status = bw_status['status']
-                if status:
-                    if status in {'WEAK', 'BREACHED'}:
-                        status = 'High-Risk Password'
-                    elif status == 'IGNORE':
-                        status = 'Ignored'
-                    print('{0:>20s}: {1:<20s}'.format('BreachWatch', status))
-
-        print('')
-
     def mask_password(self):
         if self.password != '******':
             self.unmasked_password = self.password
