@@ -15,11 +15,12 @@ import json
 import logging
 import re
 from collections import Counter
+from typing import Dict, Any
 
 from .display import bcolors
 from .params import KeeperParams
 from .record import get_totp_code
-from .subfolder import get_folder_path, find_folders, BaseFolderNode
+from . import record_types, vault
 
 
 class RecordV3:
@@ -485,6 +486,14 @@ class RecordV3:
             '$id': 'multiline',
             'type': 'multiline'
         },
+        'json': {
+            '$id': 'json',
+            'type': 'json'
+        },
+        'passkey': {
+            '$id': 'passkey',
+            'type': 'passkey'
+        },
         # 'custom: {
         #   '$id': 'custom',
         #   'type': 'custom'
@@ -509,6 +518,11 @@ class RecordV3:
             'value_description': 'multiline text',
             'value': ''  # string
         },
+        'json': {
+            'type': 'json',
+            'value_description': 'json text, only validated data persisted',
+            'value': ''  # string
+        },
         # 2021-05-06 Unused file type - removed for compatibility with web vault
         # 'file': {
         #   'type': 'file',
@@ -629,7 +643,7 @@ class RecordV3:
             'type': 'bankAccount',
             'value_description': 'bank account information',
             'value': {  # object
-                'accountType': ('Checking', 'Savings', 'Other'),
+                'accountType': ('', 'Checking', 'Savings', 'Other'),
                 'otherType': '',  # string
                 'routingNumber': '',  # string
                 'accountNumber': ''  # string (required for RT bankAccount)
@@ -644,6 +658,19 @@ class RecordV3:
                 'publicKey': '',  # string
                 'privateKey': ''  # string
             }
+        },
+        'passkey': {
+            'type': 'passkey',
+            'value_description': 'Password-less login',
+            'value': {   # object
+                'privateKey': {},     # EC private key in JsonWebKey format (RFC 7517)
+                'credentialId': '',   # 32 bytes; base64url encoded
+                'signCount': 0,       # we don't currently implement incrementing signCount
+                'userId': '',         # supplied by relyingParty at registration; base64url encoded
+                'relyingParty': '',   # utf8; i.e. "keepersecurity.com"
+                'username': '',       # user.name from webauthn create request
+                'createdDate': 0,     # unix timestamp
+            }
         }
     }
 
@@ -1573,17 +1600,9 @@ class RecordV3:
     @staticmethod
     def display(r, **kwargs):
         record_uid = r['record_uid']
-        print('')
         # print('{0:>20s}: https://keepersecurity.com/vault#detail/{1}'.format('Link', record_uid))
         print('{0:>20s}: {1:<20s}'.format('UID', record_uid))
         # if 'version' in r: print('{0:>20s}: {1:<20s}'.format('Version', str(r['version'])))
-        params = None
-        if 'params' in kwargs:
-            params = kwargs['params']
-            folders = [get_folder_path(params, x) for x in find_folders(params, record_uid)]
-            for i in range(len(folders)):
-                print('{0:>21s} {1:<20s}'.format('Folder:' if i == 0 else '', folders[i]))
-
         data = {}
         if 'data_unencrypted' in r:
             data = r['data_unencrypted'].decode() if isinstance(r['data_unencrypted'], bytes) else r['data_unencrypted']
@@ -1609,18 +1628,20 @@ class RecordV3:
             flds = c.get('value') or []
             fval = flds
             fkey = '{} ({})'.format(flab, ftyp)
-            if ftyp in ['securityQuestion', 'paymentCard', 'host', 'keyPair']:
+
+            if ftyp in ['name', 'securityQuestion', 'paymentCard', 'host', 'keyPair', 'bankAccount', 'phone']:
                 if not isinstance(flds, list):
                     flds = [flds]
                 fval = []
                 for x in flds:
                     if isinstance(x, dict):
-                        if ftyp == 'securityQuestion':
-                            q = x.get('question') or ''
-                            if q and not q.endswith('?'):
-                                q += '?'
-                            a = x.get('answer', '') if unmask else '********'
-                            fval.append(f'{q} {a}')
+                        is_empty = not any(True for k in x.values() if k)
+                        if is_empty:
+                            continue
+                        if ftyp == 'name':
+                            fval.append(vault.TypedField.export_name_field(x))
+                        elif ftyp == 'securityQuestion':
+                            fval.append(vault.TypedField.export_q_and_a_field(x))
                         elif ftyp == 'paymentCard':
                             n = x.get('cardNumber') or ''
                             if n and not unmask:
@@ -1636,23 +1657,39 @@ class RecordV3:
                             if private_key:
                                 fval.append(private_key if unmask else '********')
                         elif ftyp == 'host':
-                            hostname = x.get('hostName') or ''
-                            port = x.get('port') or ''
-                            if port:
-                                hostname += f':{port}'
-                            if hostname:
-                                fval.append(hostname)
+                            fval.append(vault.TypedField.export_host_field(x))
+                        elif ftyp == 'phone':
+                            fval.append(vault.TypedField.export_phone_field(x))
+                        elif ftyp == 'bankAccount':
+                            account_type = x.get('accountType') or ''
+                            routing_number = x.get('routingNumber') or ''
+                            if routing_number and not unmask:
+                                routing_number = '*' + routing_number[-3:]
+                            account_number = x.get('accountNumber') or ''
+                            if account_number and not unmask:
+                                account_number = '*' + account_number[-3:]
+                            number = f'{account_type}: ' if account_type else ''
+                            if routing_number or account_number:
+                                if routing_number and account_number:
+                                    number += f'{routing_number} / {account_number}'
+                                else:
+                                    number = routing_number if routing_number else account_number
+                            fval.append(number)
+
                 for i, line in enumerate(fval):
                     print('{0:>20s}: {1:<s}'.format(fkey if i == 0 else '', line))
             else:
                 if isinstance(flds, list) and len(flds) == 1:
                     if not flds[0]: continue
                     if isinstance(flds[0], dict):
-                        fval = (' ' if ftyp.lower() == 'name' else ' | ').join((str(x) for x in flds[0].values()))
+                        if (ftyp not in record_types.RecordFields or ftyp in ('passkey')) and not unmask:
+                            fval = '********'
+                        else:
+                            fval = json.dumps(flds[0], indent=2)
                     elif RecordV3.get_field_type(ftyp).get('type') == 'date' and bool(
                             re.match('^[+-]?[0-9]+$', str(flds[0]).strip())):
-                        dt = datetime.datetime.fromtimestamp(flds[0] / 1000.0)
-                        fval = str(dt)
+                        dt = datetime.datetime.fromtimestamp(int(flds[0] / 1000), tz=datetime.timezone.utc)
+                        fval = str(dt.date())
                     elif record_type == 'ssnCard' and ftyp == 'accountNumber' and flab == 'identityNumber':
                         fval = flds[0] if unmask else '********'
                     else:
@@ -1661,15 +1698,16 @@ class RecordV3:
                     if ftyp in ('fileRef', 'cardRef', 'addressRef'):
                         RecordV3.display_ref(ftyp, fval, **kwargs)
                     else:
-                        is_masked = ftyp in ['password', 'pinCode', 'secret', 'note', 'oneTimeCode'] and not unmask
+                        is_masked = (ftyp not in record_types.RecordFields or
+                                     ftyp in ['password', 'pinCode', 'secret', 'note', 'oneTimeCode']) and not unmask
                         if is_masked:
                             print('{0:>20s}: {1:<s}'.format(fkey, '********'))
                         else:
-                            if ftyp == 'multiline' and isinstance(fval, str):
+                            if isinstance(fval, str):
                                 lines = fval.split('\n')
-                                lines = [x.strip() for x in lines if x]
+                                lines = [x.rstrip() for x in lines if x]
                                 for i, line in enumerate(lines):
-                                    print('{0:>20s}: {1}'.format(fkey if i == 0 else '', line.strip()))
+                                    print('{0:>20s}: {1}'.format(fkey if i == 0 else '', line))
                             else:
                                 print('{0:>20s}: {1:<s}'.format(fkey, str(fval)))
 
@@ -1682,79 +1720,6 @@ class RecordV3:
                 code, remain, _ = result
                 if code: print('{0:>20s}: {1:<20s} valid for {2} sec'.format('Two Factor Code', code, remain))
 
-        if params is not None:
-            if record_uid in params.record_cache:
-                rec = params.record_cache[record_uid]
-                if 'shares' in rec:
-                    no = 0
-                    if 'user_permissions' in rec['shares']:
-                        perm = rec['shares']['user_permissions'].copy()
-                        perm.sort(key=lambda r: (' 1' if r.get('owner') else ' 2' if r.get(
-                            'editable') else ' 3' if r.get('sharable') else '') + r.get('username'))
-                        for uo in perm:
-                            flags = ''
-                            if uo.get('owner'):
-                                flags = 'Owner'
-                            elif uo.get('awaiting_approval'):
-                                flags = 'Awaiting Approval'
-                            else:
-                                if uo.get('editable'):
-                                    flags = 'Edit'
-                                if uo.get('sharable'):
-                                    if flags:
-                                        flags = flags + ', '
-                                    flags = flags + 'Share'
-                            if not flags:
-                                flags = 'View'
-
-                            print('{0:>21s} {1} ({2}) {3}'.format('Shared Users:' if no == 0 else '', uo['username'],
-                                                                  flags,
-                                                                  'self' if uo['username'] == params.user else ''))
-                            no = no + 1
-                    no = 0
-                    if 'shared_folder_permissions' in rec['shares']:
-                        for sfo in rec['shares']['shared_folder_permissions']:
-                            flags = ''
-                            if sfo.get('editable'):
-                                flags = 'Edit'
-                            if sfo.get('reshareable'):
-                                if flags:
-                                    flags = flags + ', '
-                                flags = flags + 'Share'
-                            if not flags:
-                                flags = 'View'
-                            sf_uid = sfo['shared_folder_uid']
-                            for f_uid in find_folders(params, record_uid):
-                                if f_uid in params.subfolder_cache:
-                                    fol = params.folder_cache[f_uid]
-                                    if fol.type in {BaseFolderNode.SharedFolderType,
-                                                    BaseFolderNode.SharedFolderFolderType}:
-                                        sfid = fol.uid if fol.type == BaseFolderNode.SharedFolderType else fol.shared_folder_uid
-                                        if sf_uid == sfid:
-                                            print('{0:>21s} {1:<20s}'.format('Shared Folders:' if no == 0 else '',
-                                                                             fol.name))
-                                            no = no + 1
-
-        if kwargs.get('format') == 'detail':
-            if 'shared' in r: print('{0:>20s}: {1:<20s}'.format('Shared', str(r['shared'])))
-            if 'client_modified_time' in r:
-                dt = datetime.datetime.fromtimestamp(r['client_modified_time'] / 1000.0)
-                print('{0:>20s}: {1:<20s}'.format('Last Modified', dt.strftime('%Y-%m-%d %H:%M:%S')))
-            if 'revision' in r: print('{0:>20s}: {1:<20s}'.format('Revision', str(r['revision'])))
-
-        if params.breach_watch:
-            bw_status = params.breach_watch.get_record_status(params, record_uid)
-            if bw_status and 'status' in bw_status:
-                status = bw_status['status']
-                if status:
-                    if status in {'WEAK', 'BREACHED'}:
-                        status = 'High-Risk Password'
-                    elif status == 'IGNORE':
-                        status = 'Ignored'
-                    print('{0:>20s}: {1:<20s}'.format('BreachWatch', status))
-
-        print('')
-
     @staticmethod
     def display_ref(ftype, fvalue, **kwargs):
         # fileRef can hold multiple references - convert single value to list for compatibility
@@ -1766,7 +1731,10 @@ class RecordV3:
                 frec = params.record_cache.get(fuid) or {}
                 fdat = frec.get('data_unencrypted') or '{}'
                 fdic = RecordV3.record_type_to_dict(fdat)
-                name = fdic.get('name') or ''
+                title = str(fdic.get('title') or '')
+                name = str(fdic.get('name') or '')
+                if title:
+                    name = title + f' ({name})'
                 size = fdic.get('size') or ''
                 if isinstance(size, str) and size.isdigit():
                     size = int(size)
@@ -1845,7 +1813,7 @@ class RecordV3:
 
             elif type(options_list) == list:
                 for c in options_list:
-                    if type(c) == dict:
+                    if isinstance(c, dict):
                         name = c.get('name')
                         value = c.get('value')
                         if name and value:
@@ -1855,6 +1823,14 @@ class RecordV3:
                             })
         return custom
 
+    @staticmethod
+    def get_title(cached_record):   # type: (Dict[str, Any]) -> str
+        if isinstance(cached_record, dict):
+            return json.loads(cached_record.get('data_unencrypted', b'{}')).get('title', '')
+        else:
+            raise TypeError('Record data must be of type dict')
+
+
 
 class HumanBytes:
     # Human-readable formatting of bytes, using binary (powers of 1024) or metric (powers of 1000) representation.
@@ -1882,111 +1858,3 @@ class HumanBytes:
             if unit != last_label:
                 num /= unit_step
         return cls.PRECISION_FORMATS[precision].format('-' if is_negative else '', num, unit)
-
-
-def init_recordv3_commands(params):
-    from .commands import commands, command_info, aliases
-    v3_commands = {}
-
-    def init_v3_commands():
-        from .commands.recordv2 import RecordAddCommand, RecordEditCommand, add_parser, edit_parser
-        global v3_commands
-        if v3_commands:
-            return
-        v3_commands = {
-            'record-type': {},
-            'record-type-info': {},
-            'add': {
-                'substitute': {
-                    'command': RecordAddCommand(),
-                    'parser': add_parser,
-                    'command_info': {add_parser.prog: add_parser.description},
-                    'alias': {'a': 'add'}
-                }
-            },
-            'edit': {
-                'substitute': {
-                    'command': RecordEditCommand(),
-                    'parser': edit_parser,
-                    'command_info': {edit_parser.prog: edit_parser.description},
-                    'alias': {}
-                }
-            }
-        }
-
-    try:
-        # parse v3 commands
-        init_v3_commands()
-
-        for cmd in v3_commands:
-            if not v3_commands[cmd].get('command') and cmd in commands:
-                prog = commands[cmd].get_parser().prog
-                parts = prog.split('|')
-                alias = parts[-1] if len(parts) == 2 else None
-                v3_commands[cmd]['command'] = {cmd: commands[cmd]}
-                if prog in command_info:
-                    v3_commands[cmd]['command_info'] = {prog: command_info[prog]}
-                if alias in aliases:
-                    v3_commands[cmd]['alias'] = {alias: aliases[alias]}
-
-        v3_enabled = params.settings.get('record_types_enabled') if params.settings and isinstance(
-            params.settings.get('record_types_enabled'), bool) else False
-        if v3_enabled:
-
-            # add/replace v3 commands
-            modified = []
-            for cmd in v3_commands:
-                cdic = v3_commands[cmd].get('command') or {}
-                cname = next(iter(cdic), None)
-                cmdc = commands.get(cmd) or 2
-                cmdv = ((v3_commands.get(cmd) or {}).get('command') or {}).get(cmd) or 3
-                if cname and (cname not in commands or cmdc != cmdv):
-                    modified.append(cname)
-                    commands[cname] = cdic[cname]
-                    ciname = next(iter(v3_commands[cmd].get('command_info') or {}), None)
-                    if ciname:
-                        command_info[ciname] = v3_commands[cmd]['command_info'][ciname]
-                    aname = next(iter(v3_commands[cmd].get('alias') or {}), None)
-                    if aname:
-                        aliases[aname] = v3_commands[cmd]['alias'][aname]
-
-            # RT activation during live session - only a full sync will pull any pre-existing v3 records
-            # full_sync = response_json.get('full_sync') or False
-            # if modified and not full_sync:
-            #     logging.warning(bcolors.WARNING + 'Record types - enabled. Please logout and login again if you have existing v3 records.' + bcolors.ENDC)
-        else:
-            # remove/replace v3 commands
-            for cmd in v3_commands:
-                cdic = v3_commands[cmd].get('command') or {}
-                cname = next(iter(cdic), None)
-                if cname and cname in commands:
-                    subs = v3_commands[cmd].get('substitute')
-                    subs = subs if isinstance(subs, dict) else {}
-                    sub_class = subs.get('command')
-                    if sub_class:
-                        commands[cname] = sub_class
-                    else:
-                        commands.pop(cname)
-
-                    ciname = next(iter(v3_commands[cmd].get('command_info') or {}), None)
-                    sub_ciname = next(iter(subs.get('command_info') or {}), None)
-                    if ciname:
-                        if ciname == sub_ciname:
-                            command_info[ciname] = subs['command_info'][sub_ciname]
-                        else:
-                            command_info.pop(ciname)
-                            if sub_ciname and subs['command_info'][sub_ciname]:
-                                command_info[sub_ciname] = subs['command_info'][sub_ciname]
-
-                    aname = next(iter(v3_commands[cmd].get('alias') or {}), None)
-                    sub_aname = next(iter(subs.get('alias') or {}), None)
-                    if aname:
-                        if aname == sub_aname:
-                            aliases[aname] = subs['alias'][sub_aname]
-                        else:
-                            aliases.pop(aname)
-                            if sub_aname and subs['alias'][sub_aname]:
-                                aliases[aname] = subs['alias'][sub_aname]
-
-    except Exception as e:
-        logging.debug(e)
