@@ -13,22 +13,18 @@ import requests
 import os
 import json
 import logging
+import ssl
 import time
 
-from typing import Union, Dict
+from typing import Union, Dict, Optional
 
 from .params import RestApiContext
-from .error import KeeperApiError, CommunicationError
+from .error import KeeperApiError, Error
 from .proto import APIRequest_pb2 as proto
 from . import crypto, utils
 from cryptography.hazmat.primitives.asymmetric import rsa, ec
 
-from . import __version__
-
-
-LEGACY_CLIENT_VERSION = 'c14.0.0'
-# CLIENT_VERSION = 'c' + __version__
-CLIENT_VERSION = 'c16.7.0'
+CLIENT_VERSION = 'c17.1.9'
 
 SERVER_PUBLIC_KEYS = {
     1: crypto.load_rsa_public_key(utils.base64_url_decode(
@@ -114,21 +110,23 @@ SERVER_PUBLIC_KEYS = {
 }   # type: Dict[int, Union[rsa.RSAPublicKey, ec.EllipticCurvePublicKey]]
 
 
-def encrypt_aes(data, key):
-    return crypto.encrypt_aes_v2(data, key)
-
-
-def decrypt_aes(data, key):
-    return crypto.decrypt_aes_v2(data, key)
+def encrypt_with_keeper_key(context, data: bytes) -> bytes:
+    key_id = context.server_key_id
+    if 1 <= key_id <= 6:
+        return crypto.encrypt_rsa(data, SERVER_PUBLIC_KEYS[key_id])
+    elif 7 <= key_id <= 17:
+        return crypto.encrypt_ec(data, SERVER_PUBLIC_KEYS[key_id])
+    else:
+        raise KeeperApiError('invalid_key_id', f'Key ID \"{key_id}\" is not valid.')
 
 
 def execute_rest(context, endpoint, payload):
-    # type: (RestApiContext, str, proto.ApiRequestPayload) -> Union[bytes, dict]
+    # type: (RestApiContext, str, proto.ApiRequestPayload) -> Optional[Union[bytes, dict]]
     if not context.transmission_key:
         context.transmission_key = os.urandom(32)
 
     if not context.server_key_id:
-        context.server_key_id = 1
+        context.server_key_id = 7
 
     run_request = True
     while run_request:
@@ -145,16 +143,30 @@ def execute_rest(context, endpoint, payload):
         api_request.publicKeyId = context.server_key_id
         api_request.locale = context.locale or 'en_US'
 
-        api_request.encryptedPayload = encrypt_aes(payload.SerializeToString(), context.transmission_key)
+        api_request.encryptedPayload = crypto.encrypt_aes_v2(payload.SerializeToString(), context.transmission_key)
 
         request_data = api_request.SerializeToString()
-        url = context.server_base + endpoint
+        if endpoint.startswith('https://'):
+            url = endpoint
+        else:
+            url = context.server_base + endpoint
 
-        logging.debug('>>> Request URL: [%s]', url)
-        rs = requests.post(url, data=request_data, headers={'Content-Type': 'application/octet-stream'},
-                           proxies=context.proxies, verify=context.certificate_check)
-        logging.debug('<<< Response Code: [%d]', rs.status_code)
-        logging.debug('<<< Response Headers: [%s]', str(rs.headers))
+        try:
+            rs = requests.post(url, data=request_data, headers={'Content-Type': 'application/octet-stream'},
+                               proxies=context.proxies, verify=context.certificate_check)
+        except requests.exceptions.SSLError as e:
+            doc_url = 'https://docs.keeper.io/secrets-manager/commander-cli/using-commander/troubleshooting-commander-cli#ssl-certificate-errors'
+            if len(e.args) > 0:
+                inner_e = e.args[0]
+                if hasattr(inner_e, 'reason'):
+                    reason = getattr(inner_e, 'reason')
+                    if isinstance(reason, Exception) and hasattr(reason, 'args'):
+                        args = getattr(reason, 'args')
+                        if isinstance(args, tuple) and len(args) > 0:
+                            inner_e = args[0]
+                            if isinstance(inner_e, ssl.SSLCertVerificationError):
+                                raise Error(f'Certificate validation error. More info:\n{doc_url}')
+            raise e
 
         content_type = rs.headers.get('Content-Type') or ''
         if rs.status_code == 200:
@@ -163,10 +175,10 @@ def execute_rest(context, endpoint, payload):
 
             rs_body = rs.content
             if rs_body:
-                rs_body = decrypt_aes(rs.content, context.transmission_key)
+                rs_body = crypto.decrypt_aes_v2(rs.content, context.transmission_key)
             return rs_body
         elif rs.status_code >= 400:
-            if content_type == 'application/json':
+            if content_type.startswith('application/json'):
                 failure = rs.json()
                 logging.debug('<<< Response Error: [%s]', failure)
                 if rs.status_code == 401:
@@ -177,101 +189,23 @@ def execute_rest(context, endpoint, payload):
                             run_request = True
                             continue
                 elif rs.status_code == 403:
-                    if failure.get('error') == 'throttled':
+                    if failure.get('error') == 'throttled' and not context.fail_on_throttle:
                         logging.info('Throttled. sleeping for 10 seconds')
                         time.sleep(10)
                         run_request = True
                         continue
                 return failure
             else:
-                logging.debug('<<< HTTP Status: [%s]  Reason: [%s]', rs.status_code, rs.reason)
                 if logging.getLogger().level <= logging.DEBUG:
                     if rs.text:
-                        logging.debug('<<< Response Content: [%s]', str(rs.text))
-                raise CommunicationError('Code {0}: {1}'.format(rs.status_code, rs.reason))
-
-
-def get_device_token(context):
-    # type: (RestApiContext) -> str
-
-    if not context.device_id:
-        rq = proto.DeviceRequest()
-        rq.clientVersion = CLIENT_VERSION
-        rq.deviceName = ''
-
-        api_request_payload = proto.ApiRequestPayload()
-        api_request_payload.payload = rq.SerializeToString()
-        rs = execute_rest(context, 'authentication/get_device_token', api_request_payload)
-        if type(rs) is bytes:
-            device_rs = proto.DeviceResponse()
-            device_rs.ParseFromString(rs)
-            if proto.DeviceStatus.Name(device_rs.status) == 'DEVICE_OK':
-                context.device_id = device_rs.encryptedDeviceToken
-        elif type(rs) is dict:
-            raise KeeperApiError(rs['error'], rs['message'])
-
-    return context.device_id
-
-
-def pre_login(context, username, two_factor_token=None):
-    # type: (RestApiContext, str, bytes or None) -> proto.PreLoginResponse
-
-    attempt = 0
-    while attempt < 3:
-        attempt += 1
-        rq = proto.PreLoginRequest()
-        rq.authRequest.clientVersion = CLIENT_VERSION
-        rq.authRequest.username = username.lower()
-        rq.authRequest.encryptedDeviceToken = get_device_token(context)
-        rq.loginType = proto.LoginType.Value('NORMAL')
-        if two_factor_token:
-            rq.twoFactorToken = two_factor_token
-
-        api_request_payload = proto.ApiRequestPayload()
-        api_request_payload.payload = rq.SerializeToString()
-        rs = execute_rest(context, 'authentication/pre_login', api_request_payload)
-        if type(rs) == bytes:
-            pre_login_rs = proto.PreLoginResponse()
-            pre_login_rs.ParseFromString(rs)
-            return pre_login_rs
-
-        elif type(rs) is dict:
-            if 'error' in rs and 'message' in rs:
-                if rs['error'] == 'region_redirect':
-                    context.device_id = None
-                    context.server_base = 'https://{0}/'.format(rs['region_host'])
-                    logging.warning('Switching to region: %s', rs['region_host'])
-                    continue
-                if rs['error'] == 'bad_request':
-                    logging.warning('Pre-Auth error: %s', rs.get('additional_info'))
-                    context.device_id = None
-                    continue
-
-                raise KeeperApiError(rs['error'], rs['message'])
-    raise CommunicationError('Cannot get user information')
-
-
-def get_new_user_params(context, username):
-    # type: (RestApiContext, str) -> proto.NewUserMinimumParams
-    rq = proto.AuthRequest()
-    rq.clientVersion = CLIENT_VERSION
-    rq.username = username.lower()
-    rq.encryptedDeviceToken = get_device_token(context)
-
-    api_request_payload = proto.ApiRequestPayload()
-    api_request_payload.payload = rq.SerializeToString()
-    rs = execute_rest(context, 'authentication/get_new_user_params', api_request_payload)
-    if type(rs) is bytes:
-        pre_login_rs = proto.NewUserMinimumParams()
-        pre_login_rs.ParseFromString(rs)
-        return pre_login_rs
-
-    if type(rs) is dict:
-        raise KeeperApiError(rs['error'], rs['message'])
+                        logging.debug('<<< Response Content: [%s]', rs.text)
+                    else:
+                        logging.debug('<<< HTTP Status: [%s]  Reason: [%s]', rs.status_code, rs.reason)
+                raise KeeperApiError(rs.status_code, rs.reason)
 
 
 def v2_execute(context, rq):
-    # type: (RestApiContext, dict) -> dict
+    # type: (RestApiContext, dict) -> Optional[dict]
 
     api_request_payload = proto.ApiRequestPayload()
     api_request_payload.payload = json.dumps(rq).encode('utf-8')
