@@ -11,52 +11,89 @@
 import abc
 import itertools
 import re
-from typing import Optional, Union, Iterator, Dict, Set
+from typing import Optional, Union, Iterator, Dict, Set, Callable, Any, Iterable
 
 from . import crypto, utils, vault, record_types
 from .params import KeeperParams
 
 
-def find_records(params, search_str=None, record_type=None):
-    # type: (KeeperParams, Optional[str], Optional[Union[str, Iterator[str]]]) -> Iterator[vault.KeeperRecord]
-    pattern = re.compile(search_str, re.IGNORECASE) if search_str else None
+def _match_value(pattern, value):  # type: (Callable[[str], Any], Any) -> bool
+    if isinstance(value, str):
+        if pattern(value):
+            return True
+    elif isinstance(value, list):
+        for v in value:
+            if _match_value(pattern, v):
+                return True
+    elif isinstance(value, dict):
+        for v in value.values():
+            if _match_value(pattern, v):
+                return True
+    return False
 
-    type_filter = None
+
+def matches_record(record, pattern, search_fields=None):    # type: (vault.KeeperRecord, Union[str, Callable[[str], Any]], Optional[Iterable[str]]) -> bool
+    if isinstance(pattern, str):
+        pattern = re.compile(pattern, re.IGNORECASE).search
+
+    if search_fields is not None:
+        search_fields = {f.lower() for f in search_fields}
+
+    for key, value in record.enumerate_fields():
+        m = re.match(r'^\((\w+)\)\.?', key)
+        if m:
+            key = m.group(1)
+        if search_fields is not None and key.lower() not in search_fields:
+            continue
+        if key and _match_value(pattern, key):
+            return True
+        if value and _match_value(pattern, value):
+            return True
+    return False
+
+
+def find_records(params,                   # type: KeeperParams
+                 search_str=None,          # type: Optional[str]
+                 record_type=None,         # type: Union[str, Iterable[str], None]
+                 record_version=None,      # type: Union[int, Iterable[int], None]
+                 search_fields=None        # type: Optional[Iterable[str]]
+                 ):                       # type: (...) -> Iterator[vault.KeeperRecord]
+    pattern = re.compile(search_str, re.IGNORECASE).search if search_str else None
+
+    type_filter = None       # type: Optional[Set[str]]
+    version_filter = None    # type: Optional[Set[int]]
     if record_type:
         type_filter = set()
-        if hasattr(record_type, '__iter__'):
-            type_filter.update(record_type)
-        elif isinstance(record_type, str):
+        if isinstance(record_type, str):
             type_filter.add(record_type)
+        if isinstance(record_type, Iterable):
+            type_filter.update(record_type)
+
+    if record_version:
+        version_filter = set()
+        if isinstance(record_version, int):
+            version_filter.add(record_version)
+        if isinstance(record_version, Iterable):
+            version_filter.update((x for x in record_version if isinstance(x, int)))
 
     for record_uid in params.record_cache:
         record = vault.KeeperRecord.load(params, record_uid)
         if not record:
             continue
+        if search_str and record.record_uid == search_str:
+            yield record
+            continue
+        if version_filter and record.version not in version_filter:
+            continue
         if type_filter and record.record_type not in type_filter:
             continue
-        if not pattern:
-            yield record
+
+        if pattern:
+            is_match = matches_record(record, pattern, search_fields)
         else:
-            is_match = search_str == record.record_uid
-            if not is_match and not type_filter:
-                is_match = pattern.search(record.record_type)
-            if not is_match:
-                for _, value in record.enumerate_fields():
-                    if not value:
-                        continue
-                    if is_match:
-                        break
-                    if isinstance(value, str):
-                        is_match = pattern.search(value)
-                    elif isinstance(value, list):
-                        for token in value:
-                            if is_match:
-                                break
-                            if isinstance(token, str):
-                                is_match = pattern.search(token)
-            if is_match:
-                yield record
+            is_match = True
+        if is_match:
+            yield record
 
 
 def get_record_description(record):   # type: (vault.KeeperRecord) -> Optional[str]
@@ -67,31 +104,71 @@ def get_record_description(record):   # type: (vault.KeeperRecord) -> Optional[s
         return ' @ '.join((str(x) for x in comps if x))
 
     if isinstance(record, vault.TypedRecord):
+        if record.version == 6:
+            return 'PAM Configuration'
         field = next((x for x in record.fields if x.type == 'login'), None)
         if field:
-            comps.append(field.get_default_value() or '')
-            field = next((x for x in record.fields if x.type == 'url'), None)
-            if field:
-                comps.append(field.get_default_value())
-            else:
-                field = next((x for x in record.fields if x.type == 'host'), None)
+            value = field.get_default_value()
+            if value:
+                comps.append(field.get_default_value() or '')
+                field = next((x for x in record.fields if x.type == 'url'), None)
                 if field:
-                    host = field.get_default_value()
-                    if isinstance(host, dict):
-                        address = host.get('hostName')
-                        if address:
-                            port = host.get('port')
-                            if port:
-                                address = f'{address}:{port}'
-                            comps.append(address)
-            return ' @ '.join((str(x) for x in comps if x))
+                    comps.append(field.get_default_value())
+                else:
+                    field = next((x for x in itertools.chain(record.fields, record.custom) if x.type.lower() in {'host', 'pamhostname'}), None)
+                    if field:
+                        host = field.get_default_value()
+                        if isinstance(host, dict):
+                            address = host.get('hostName')
+                            if address:
+                                port = host.get('port')
+                                if port:
+                                    address = f'{address}:{port}'
+                                comps.append(address)
+                return ' @ '.join((str(x) for x in comps if x))
 
-        field = next((x for x in record.fields if x.type == 'name'), None)
+        field = next((x for x in record.fields if x.type == 'paymentCard'), None)
         if field:
             value = field.get_default_value()
             if isinstance(value, dict):
-                comps.extend((value.get('first', ''), value.get('middle', ''), value.get('last', '')))
-            return ' '.join((str(x) for x in comps if x))
+                number = value.get('cardNumber') or ''
+                if isinstance(number, str):
+                    if len(number) > 4:
+                        number = '*' + number[-4:]
+                        comps.append(number)
+
+                field = next((x for x in record.fields if x.type == 'text' and x.label == 'cardholderName'), None)
+                if field:
+                    name = field.get_default_value()
+                    if name and isinstance(name, str):
+                        comps.append(name.upper())
+                return ' / '.join((str(x) for x in comps if x))
+
+        field = next((x for x in record.fields if x.type == 'bankAccount'), None)
+        if field:
+            value = field.get_default_value()
+            if isinstance(value, dict):
+                routing = value.get('routingNumber') or ''
+                if routing:
+                    routing = '*' + routing[-3:]
+                account = value.get('accountNumber') or ''
+                if account:
+                    account = '*' + account[-3:]
+                if routing or account:
+                    if routing and account:
+                        return f'{routing} / {account}'
+                    else:
+                        return routing if routing else account
+
+        field = next((x for x in record.fields if x.type == 'keyPair'), None)
+        if field:
+            value = field.get_default_value()
+            if isinstance(value, dict):
+                if value.get('privateKey'):
+                    comps.append('<Private Key>')
+                if value.get('publicKey'):
+                    comps.append('<Public Key>')
+            return ' / '.join((str(x) for x in comps if x))
 
         field = next((x for x in record.fields if x.type == 'address'), None)
         if field:
@@ -104,32 +181,12 @@ def get_record_description(record):   # type: (vault.KeeperRecord) -> Optional[s
                     f'{value.get("country", "")}'))
             return ', '.join((str(x) for x in comps if x))
 
-        field = next((x for x in record.fields if x.type == 'paymentCard'), None)
+        field = next((x for x in record.fields if x.type == 'name'), None)
         if field:
             value = field.get_default_value()
             if isinstance(value, dict):
-                number = value.get('cardNumber', '')
-                if isinstance(number, str):
-                    if len(number) > 4:
-                        number = 'x' + number[-4:]
-                        comps.append(number)
-
-            field = next((x for x in record.fields if x.type == 'text' and x.label == 'cardholderName'), None)
-            if field:
-                name = field.get_default_value()
-                if name and isinstance(name, str):
-                    comps.append(name.upper())
-            return ' / '.join((str(x) for x in comps if x))
-
-        field = next((x for x in record.fields if x.type == 'keyPair'), None)
-        if field:
-            value = field.get_default_value()
-            if isinstance(value, dict):
-                if value.get('privateKey'):
-                    comps.append('<Private Key>')
-                if value.get('publicKey'):
-                    comps.append('<Public Key>')
-            return ' / '.join((str(x) for x in comps if x))
+                comps.extend((value.get('first', ''), value.get('middle', ''), value.get('last', '')))
+                return ' '.join((str(x) for x in comps if x))
 
     if isinstance(record, vault.FileRecord):
         comps.extend((record.name, utils.size_to_str(record.size)))
@@ -185,7 +242,7 @@ def extract_password_record_extras(record, existing_extra=None):
                     'field_title': ''
                 }
                 extra['fields'].append(totp_field)
-            totp_field['data'] = totp_field
+            totp_field['data'] = record.totp
         else:
             if totp_field:
                 extra['fields'].remove(totp_field)
@@ -249,11 +306,14 @@ def extract_typed_field(field):     # type: (vault.TypedField) -> dict
                         if key not in value:
                             value[key] = ''
                 field_values.append(value)
-    return {
+    result = {
         'type': field_type,
         'label': field.label or '',
         'value': field_values
     }
+    if field.required is True:
+        result['required'] = True
+    return result
 
 
 def extract_typed_record_data(record):      # type: (vault.TypedRecord) -> dict
@@ -274,11 +334,23 @@ def extract_typed_record_data(record):      # type: (vault.TypedRecord) -> dict
 def extract_typed_record_refs(record):  # type: (vault.TypedRecord) -> Set[str]
     refs = set()
     for field in itertools.chain(record.fields, record.custom):
-        if field.type in {'fileRef', 'addressRef', 'cardRef'}:
-            if isinstance(field.value, list):
-                for ref in refs:
+        if field.type in {'fileRef', 'addressRef', 'cardRef', 'recordRef'}:
+            if not isinstance(field.value, list):
+                continue
+            for ref in field.value:
+                if isinstance(ref, str):
+                    refs.add(ref)
+        elif field.type == 'script':
+            if not isinstance(field.value, list):
+                continue
+            for script in field.value:
+                if not isinstance(script, dict):
+                    continue
+                if 'fileRef' in script:
+                    ref = script['fileRef']
                     if isinstance(ref, str):
                         refs.add(ref)
+
     return refs
 
 
@@ -291,13 +363,54 @@ class TypedRecordFacade(abc.ABC):
         pass
 
     def assign_record(self, record):  # type: (vault.TypedRecord) -> None
-        if self._get_facade_type() != record.record_type:
-            raise Exception(f'Incorrect record type: expected {self._get_facade_type()}, got {record.record_type}')
+        if not record.record_type:
+            record.type_name = self._get_facade_type()
         self.record = record
 
     def get_custom_field(self, name):
         return next((x.get_default_value(str) for x in self.record.custom if x.label.lower() == name.lower()), '')
 
+    @property
+    def title(self):  # type: () -> Optional[str]
+        if self.record:
+            return self.record.title
+
+    @title.setter
+    def title(self, value):   # type: (str) -> None
+        if self.record:
+            self.record.title = value
+
+    @property
+    def notes(self):  # type: () -> Optional[str]
+        if self.record:
+            return self.record.notes
+
+    @notes.setter
+    def notes(self, value):   # type: (str) -> None
+        if self.record:
+            self.record.notes = value
+
+
+class FileRefFacade(TypedRecordFacade, abc.ABC):
+    def __init__(self):
+        TypedRecordFacade.__init__(self)
+        self._file_ref_field = None   # type: Optional[vault.TypedField]
+
+    def assign_record(self, record):
+        super(FileRefFacade, self).assign_record(record)
+        if isinstance(record, vault.TypedRecord):
+            self._file_ref_field = self.record.get_typed_field('fileRef')
+            if self._file_ref_field is None:
+                self._file_ref_field = vault.TypedField.new_field('fileRef', [])
+                record.fields.append(self._file_ref_field)
+        else:
+            self._file_ref_field = None
+
+    @property
+    def file_ref(self):
+        if self._file_ref_field:
+            return self._file_ref_field.value
+
 
 class ServerFacade(TypedRecordFacade, abc.ABC):
     def __init__(self):
@@ -406,3 +519,76 @@ class DatabaseCredentialsFacade(ServerCredentialsFacade):
     def database_type(self):
         if self.database_type_field:
             return self.database_type_field.get_default_value(value_type=str)
+
+
+class LoginFacade(FileRefFacade, abc.ABC):
+    def __init__(self):
+        FileRefFacade.__init__(self)
+        self._login_field = None      # type: Optional[vault.TypedField]
+        self._password_field = None   # type: Optional[vault.TypedField]
+        self._url = None              # type: Optional[vault.TypedField]
+
+    def assign_record(self, record):
+        super(FileRefFacade, self).assign_record(record)
+        if record:
+            self._login_field = self.record.get_typed_field('login')
+            if self._login_field is None:
+                self._login_field = vault.TypedField.new_field('login', '')
+                record.fields.append(self._login_field)
+            self._password_field = self.record.get_typed_field('password')
+            if self._password_field is None:
+                self._password_field = vault.TypedField.new_field('password', '')
+                record.fields.append(self._password_field)
+            self._url = self.record.get_typed_field('url')
+            if self._url is None:
+                self._url = vault.TypedField.new_field('url', '')
+                record.fields.append(self._url)
+        else:
+            self._login_field = None
+            self._password_field = None
+            self._url = None
+
+    def _get_facade_type(self):
+        return 'login'
+
+    @property
+    def login(self):
+        return self._login_field.get_default_value(str) if self._login_field else ''
+
+    @login.setter
+    def login(self, value):
+        if self._login_field:
+            holder = self._login_field.value
+            if isinstance(holder, list):
+                if len(holder) > 0:
+                    holder[0] = value
+                else:
+                    holder.append(value)
+
+    @property
+    def password(self):
+        return self._password_field.get_default_value(str) if self._password_field else ''
+
+    @password.setter
+    def password(self, value):
+        if self._password_field:
+            holder = self._password_field.value
+            if isinstance(holder, list):
+                if len(holder) > 0:
+                    holder[0] = value
+                else:
+                    holder.append(value)
+
+    @property
+    def url(self):
+        return self._url.get_default_value(str) if self._url else ''
+
+    @url.setter
+    def url(self, value):
+        if self._url:
+            holder = self._url.value
+            if isinstance(holder, list):
+                if len(holder) > 0:
+                    holder[0] = value
+                else:
+                    holder.append(value)
